from typing import List

from matplotlib.artist import Artist
from matplotlib.axes import Axes

from newclid.all_rules import (
    R03_ARC_DETERMINES_INTERNAL_ANGLES,
    R04_CONGRUENT_ANGLES_ARE_IN_A_CYCLIC,
    R11_BISECTOR_THEOREM_1,
    R12_BISECTOR_THEOREM_2,
    R13_ISOSCELES_TRIANGLE_EQUAL_ANGLES,
    R14_EQUAL_BASE_ANGLES_IMPLY_ISOSCELES,
    R19_HYPOTENUSE_IS_DIAMETER,
    R28_OVERLAPPING_PARALLELS,
    R34_AA_SIMILARITY_OF_TRIANGLES_DIRECT,
    R35_AA_SIMILARITY_OF_TRIANGLES_REVERSE,
    R41_THALES_THEOREM_3,
    R42_THALES_THEOREM_4,
    R43_ORTHOCENTER_THEOREM,
    R46_INCENTER_THEOREM,
    R49_RECOGNIZE_CENTER_OF_CIRCLE_CYCLIC,
    R51_MIDPOINT_SPLITS_IN_TWO,
    R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES,
    R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES,
    R54_DEFINITION_OF_MIDPOINT,
    R56_MIDPOINT_COLL_PROPERTIES,
    R60_SSS_SIMILARITY_OF_TRIANGLES_DIRECT,
    R61_SSS_SIMILARITY_OF_TRIANGLES_REVERSE,
    R62_SAS_SIMILARITY_OF_TRIANGLES_DIRECT,
    R63_SAS_SIMILARITY_OF_TRIANGLES_REVERSE,
    R68_SIMILARITY_WITHOUT_SCALING_DIRECT,
    R69_SIMILARITY_WITHOUT_SCALING_REVERSE,
    R71_RESOLUTION_OF_RATIOS,
    R72_DISASSEMBLING_A_CIRCLE,
    R74_INTERSECTION_BISECTORS,
    R82_PARA_OF_COLL,
)
from newclid.draw.geometries import draw_arrow, draw_circle, draw_segment, draw_triangle
from newclid.draw.predicates import (
    draw_free_perpendicular_symbol,
    draw_line,
    draw_predicate,
)
from newclid.draw.theme import DrawTheme
from newclid.jgex.geometries import (
    JGEXPoint,
    line_line_intersection,
    perpendicular_bisector,
)
from newclid.justifications.justification import RuleApplication
from newclid.numerical.geometries import PointNum
from newclid.predicates import Predicate
from newclid.predicates._index import PredicateType
from newclid.predicates.circumcenter import Circumcenter
from newclid.predicates.collinearity import Coll
from newclid.predicates.equal_angles import EqAngle
from newclid.predicates.triangles_similar import SimtriClock, SimtriReflect
from newclid.symbols.points_registry import Point, Segment
from newclid.symbols.symbols_registry import SymbolsRegistry


def draw_rule_application(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    rule_applied = application.rule

    match rule_applied.id:
        case R03_ARC_DETERMINES_INTERNAL_ANGLES.id:
            return _draw_arc_determines_internal_angles(ax, application, theme)
        case R04_CONGRUENT_ANGLES_ARE_IN_A_CYCLIC.id:
            return _draw_congruent_angles_are_in_a_cyclic(ax, application, theme)
        case R11_BISECTOR_THEOREM_1.id:
            return _draw_bisector_from_eqratio_configuration(
                ax, application, symbols, theme
            )
        case R12_BISECTOR_THEOREM_2.id:
            return _draw_bisector_gives_eqratio_configuration(
                ax, application, symbols, theme
            )
        case R13_ISOSCELES_TRIANGLE_EQUAL_ANGLES.id:
            return _draw_isosceles_triangle_angles(ax, application, symbols, theme)
        case R14_EQUAL_BASE_ANGLES_IMPLY_ISOSCELES.id:
            return _draw_isosceles_triangle_angles_reverse(
                ax, application, symbols, theme
            )
        case R19_HYPOTENUSE_IS_DIAMETER.id:
            return _draw_hypotenuse_is_diameter(ax, application, theme)
        case R28_OVERLAPPING_PARALLELS.id:
            return _draw_coll_of_para(ax, application, theme)
        case R34_AA_SIMILARITY_OF_TRIANGLES_DIRECT.id:
            return _draw_sas_similarity_of_triangles_direct(
                ax, application, symbols, theme
            )
        case R35_AA_SIMILARITY_OF_TRIANGLES_REVERSE.id:
            return _draw_sas_similarity_of_triangles_reverse(
                ax, application, symbols, theme
            )
        case R41_THALES_THEOREM_3.id:
            return _draw_thales_configuration(ax, application, symbols, theme)
        case R42_THALES_THEOREM_4.id:
            return _draw_thales_configuration_giving_eqratio(
                ax, application, symbols, theme
            )
        case R43_ORTHOCENTER_THEOREM.id:
            return _draw_orthocenter_theorem(ax, application, symbols, theme)
        case R46_INCENTER_THEOREM.id:
            return _draw_incenter_theorem(ax, application, symbols, theme)
        case R49_RECOGNIZE_CENTER_OF_CIRCLE_CYCLIC.id:
            return _draw_recognize_center_of_circle(ax, application, theme)
        case R51_MIDPOINT_SPLITS_IN_TWO.id:
            return _draw_midpoint(ax, application, theme)
        case R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES.id:
            return _draw_similar_triangles_direct_properties(ax, application, theme)
        case R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES.id:
            return _draw_similar_triangles_reverse_properties(ax, application, theme)
        case R54_DEFINITION_OF_MIDPOINT.id:
            return _draw_midpoint_consequence(ax, application, theme)
        case R56_MIDPOINT_COLL_PROPERTIES.id:
            return _draw_midpoint(ax, application, theme)
        case R60_SSS_SIMILARITY_OF_TRIANGLES_DIRECT.id:
            return _draw_sas_similarity_of_triangles_direct(
                ax, application, symbols, theme
            )
        case R61_SSS_SIMILARITY_OF_TRIANGLES_REVERSE.id:
            return _draw_sas_similarity_of_triangles_reverse(
                ax, application, symbols, theme
            )
        case R62_SAS_SIMILARITY_OF_TRIANGLES_DIRECT.id:
            return _draw_sas_similarity_of_triangles_direct(
                ax, application, symbols, theme
            )
        case R63_SAS_SIMILARITY_OF_TRIANGLES_REVERSE.id:
            return _draw_sas_similarity_of_triangles_reverse(
                ax, application, symbols, theme
            )
        case R68_SIMILARITY_WITHOUT_SCALING_DIRECT.id:
            return _draw_direct_congruence(ax, application, symbols, theme)
        case R69_SIMILARITY_WITHOUT_SCALING_REVERSE.id:
            return _draw_reverse_congruence(ax, application, symbols, theme)
        case R71_RESOLUTION_OF_RATIOS.id:
            return _draw_resolution_of_ratios(ax, application, symbols, theme)
        case R72_DISASSEMBLING_A_CIRCLE.id:
            return _draw_disassembling_a_circle(ax, application, symbols, theme)
        case R74_INTERSECTION_BISECTORS.id:
            return _draw_intersection_bisectors(ax, application, theme)
        case R82_PARA_OF_COLL.id:
            return _draw_para_of_coll(ax, application, theme)
        case _:
            return []


def _draw_arc_determines_internal_angles(
    ax: Axes, application: RuleApplication, theme: DrawTheme
) -> list[Artist]:
    if application.predicate.predicate_type != PredicateType.EQUAL_ANGLES:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    if (
        len(application.premises) < 1
        or application.premises[0].predicate_type != PredicateType.CYCLIC
    ):
        raise ValueError(
            f"Unexpected premises for rule {application.rule}: {application.premises}"
        )

    angle1 = application.predicate.angle1
    angle2 = application.predicate.angle2

    vertex1 = _one_angle_vertex(angle1)
    vertex2 = _one_angle_vertex(angle2)

    angle_edges: list[Point] = []
    for line in angle1:
        for point in line:
            if point != vertex1:
                angle_edges.append(point)

    p1 = angle_edges[0]
    p2 = angle_edges[1]

    center = _circumcenter_of_triangle((p1, p2, vertex1))
    radius = center.distance(p1.num)

    return [
        draw_circle(
            ax,
            (center.x, center.y),
            radius,
            line_color=theme.circle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex1.num,
            p1.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex1.num,
            p2.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex2.num,
            p1.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex2.num,
            p2.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
    ]


def _draw_congruent_angles_are_in_a_cyclic(
    ax: Axes,
    application: RuleApplication,
    theme: DrawTheme,
) -> list[Artist]:
    if application.predicate.predicate_type != PredicateType.CYCLIC:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    if (
        len(application.premises) < 1
        or application.premises[0].predicate_type != PredicateType.EQUAL_ANGLES
    ):
        raise ValueError(
            f"Unexpected premises for rule {application.rule}: {application.premises}"
        )

    angle1 = application.premises[0].angle1
    angle2 = application.premises[0].angle2

    vertex1 = _one_angle_vertex(angle1)
    vertex2 = _one_angle_vertex(angle2)

    angle_edges: list[Point] = []
    for line in angle1:
        for point in line:
            if point != vertex1:
                angle_edges.append(point)

    p1 = angle_edges[0]
    p2 = angle_edges[1]

    center = _circumcenter_of_triangle((p1, p2, vertex1))
    radius = center.distance(p1.num)

    return [
        draw_circle(
            ax,
            (center.x, center.y),
            radius,
            line_color=theme.circle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex1.num,
            p1.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex1.num,
            p2.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex2.num,
            p1.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex2.num,
            p2.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
    ]


def _draw_bisector_from_eqratio_configuration(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    if len(application.premises) != 3:
        raise ValueError(
            f"Unexpected number of premises for rule {application.rule}: {application.premises}"
        )
    coll = application.premises[0]
    if coll.predicate_type != PredicateType.COLLINEAR:
        raise ValueError(
            f"Unexpected premise for rule {application.rule}: {application.premises[1]}"
        )
    p1, p2, p3 = coll.points
    eqangle = application.predicate
    if eqangle.predicate_type != PredicateType.EQUAL_ANGLES:
        raise ValueError(
            f"Unexpected premise for rule {application.rule}: {application.predicate}"
        )
    vertex = _get_eqangle_vertex(eqangle)
    return [
        draw_segment(
            ax,
            p1.num,
            p2.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            p2.num,
            p3.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex.num,
            p1.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex.num,
            p2.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex.num,
            p3.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
    ] + list(draw_predicate(ax, eqangle, symbols, theme=theme))


def _draw_bisector_gives_eqratio_configuration(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    coll: Coll | None = None
    for premise in application.premises:
        if premise.predicate_type == PredicateType.COLLINEAR:
            coll = premise
            break

    if coll is None:
        raise ValueError(
            f"Missing coll premise for rule {application.rule}: {application.premises}"
        )
    p1, p2, p3 = coll.points

    eqangle: EqAngle | None = None
    for premise in application.premises:
        if premise.predicate_type == PredicateType.EQUAL_ANGLES:
            eqangle = premise
            break

    if eqangle is None:
        raise ValueError(
            f"Missing eqangle premise for rule {application.rule}: {application.premises}"
        )

    vertex = _get_eqangle_vertex(eqangle)
    return [
        draw_segment(
            ax,
            p1.num,
            p2.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            p2.num,
            p3.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex.num,
            p1.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex.num,
            p2.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            vertex.num,
            p3.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
    ] + list(draw_predicate(ax, eqangle, symbols, theme=theme))


def _draw_isosceles_triangle_angles(
    ax: Axes, application: RuleApplication, symbols: SymbolsRegistry, theme: DrawTheme
) -> list[Artist]:
    vertices: List[Point] = []
    if len(application.premises) != 1:
        raise ValueError(
            f"Unexpected number of premises for rule {application.rule}: {application.premises}"
        )
    if application.predicate.predicate_type != PredicateType.EQUAL_ANGLES:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    eqangle = application.predicate

    for point in eqangle.angle1[0]:
        if point not in vertices:
            vertices.append(point)
    for point in eqangle.angle1[1]:
        if point not in vertices:
            vertices.append(point)

    return [
        draw_triangle(
            ax,
            vertices[0].num,
            vertices[1].num,
            vertices[2].num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
    ] + list(draw_predicate(ax, eqangle, symbols, theme=theme))


def _draw_isosceles_triangle_angles_reverse(
    ax: Axes, application: RuleApplication, symbols: SymbolsRegistry, theme: DrawTheme
) -> list[Artist]:
    vertices: List[Point] = []
    if len(application.premises) != 2:
        raise ValueError(
            f"Unexpected number of premises for rule {application.rule}: {application.premises}"
        )
    if application.premises[0].predicate_type != PredicateType.EQUAL_ANGLES:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    eqangle = application.premises[0]

    for point in eqangle.angle1[0]:
        if point not in vertices:
            vertices.append(point)
    for point in eqangle.angle1[1]:
        if point not in vertices:
            vertices.append(point)

    return list(draw_predicate(ax, eqangle, symbols, theme=theme)) + [
        draw_triangle(
            ax,
            vertices[0].num,
            vertices[1].num,
            vertices[2].num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
    ]


def _draw_hypotenuse_is_diameter(
    ax: Axes, application: RuleApplication, theme: DrawTheme
) -> list[Artist]:
    vertices: List[Point] = []
    corners: List[Point] = []
    if len(application.premises) != 2:
        raise ValueError(
            f"Unexpected number of premises for rule {application.rule}: {application.premises}"
        )
    if application.premises[0].predicate_type != PredicateType.MIDPOINT:
        raise ValueError(
            f"Unexpected premise for rule {application.rule}: {application.premises[0]}"
        )
    midpoint = application.premises[0].midpoint
    if application.premises[1].predicate_type != PredicateType.PERPENDICULAR:
        raise ValueError(
            f"Unexpected premise for rule {application.rule}: {application.premises[1]}"
        )
    for point in application.premises[1].line1:
        if point not in vertices:
            vertices.append(point)
    for point in application.premises[1].line2:
        if point not in vertices:
            vertices.append(point)
    if application.predicate.predicate_type != PredicateType.CONGRUENT:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    for point in application.predicate.segment1:
        if point != midpoint:
            corners.append(point)
    for point in application.predicate.segment2:
        if point != midpoint:
            corners.append(point)
    radius = corners[0].num.distance(midpoint.num)

    return [
        draw_triangle(
            ax,
            vertices[0].num,
            vertices[1].num,
            vertices[2].num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            midpoint.num,
            corners[0].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            midpoint.num,
            corners[1].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_circle(
            ax,
            (midpoint.num.x, midpoint.num.y),
            radius,
            line_color=theme.circle_color,
            line_width=theme.thick_line_width,
        ),
    ]


def _draw_coll_of_para(
    ax: Axes,
    application: RuleApplication,
    theme: DrawTheme,
) -> list[Artist]:
    if application.predicate.predicate_type != PredicateType.COLLINEAR:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    return [
        draw_segment(
            ax,
            application.predicate.points[0].num,
            application.predicate.points[1].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            application.predicate.points[1].num,
            application.predicate.points[2].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
    ]


def _draw_para_of_coll(
    ax: Axes,
    application: RuleApplication,
    theme: DrawTheme,
) -> list[Artist]:
    coll = application.premises[0]
    if coll.predicate_type != PredicateType.COLLINEAR:
        raise ValueError(f"Unexpected premises for rule {application.rule}: {coll}")
    return [
        draw_segment(
            ax,
            coll.points[0].num,
            coll.points[1].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            coll.points[1].num,
            coll.points[2].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
    ]


def _draw_thales_configuration(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    if application.predicate.predicate_type != PredicateType.PARALLEL:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    para_consequence = application.predicate
    if len(application.premises) != 5:
        raise ValueError(
            f"Unexpected number of premises for rule {application.rule}: {application.premises}"
        )
    if application.premises[1].predicate_type != PredicateType.COLLINEAR:
        raise ValueError(
            f"Unexpected premise 1 for rule {application.rule}: {application.premises[1]}"
        )
    coll1 = application.premises[1]
    p1, p2, p3 = coll1.points
    if application.premises[2].predicate_type != PredicateType.EQUAL_RATIOS:
        raise ValueError(
            f"Unexpected premise 2 for rule {application.rule}: {application.premises[2]}"
        )
    eqratio = application.premises[2]
    if application.premises[3].predicate_type != PredicateType.PARALLEL:
        raise ValueError(
            f"Unexpected premise 3 for rule {application.rule}: {application.premises[3]}"
        )
    para_premise = application.premises[3]

    if application.premises[0].predicate_type != PredicateType.COLLINEAR:
        raise ValueError(
            f"Unexpected premise 0 for rule {application.rule}: {application.premises[0]}"
        )
    coll2 = application.premises[0]
    q1, q2, q3 = coll2.points

    return (
        list(draw_predicate(ax, para_consequence, symbols, theme=theme))
        + list(draw_predicate(ax, eqratio, symbols, theme=theme))
        + list(draw_predicate(ax, para_premise, symbols, theme=theme))
        + [
            draw_segment(
                ax,
                p1.num,
                p2.num,
                line_color=theme.line_color,
                line_width=theme.thick_line_width,
            ),
            draw_segment(
                ax,
                p2.num,
                p3.num,
                line_color=theme.line_color,
                line_width=theme.thick_line_width,
            ),
            draw_segment(
                ax,
                q1.num,
                q2.num,
                line_color=theme.line_color,
                line_width=theme.thick_line_width,
            ),
            draw_segment(
                ax,
                q2.num,
                q3.num,
                line_color=theme.line_color,
                line_width=theme.thick_line_width,
            ),
        ]
    )


def _draw_thales_configuration_giving_eqratio(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    if application.premises[2].predicate_type != PredicateType.PARALLEL:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    para_consequence = application.premises[2]
    if len(application.premises) != 4:
        raise ValueError(
            f"Unexpected number of premises for rule {application.rule}: {application.premises}"
        )
    if application.premises[1].predicate_type != PredicateType.COLLINEAR:
        raise ValueError(
            f"Unexpected premise 1 for rule {application.rule}: {application.premises[1]}"
        )
    coll1 = application.premises[1]
    p1, p2, p3 = coll1.points
    if application.predicate.predicate_type != PredicateType.EQUAL_RATIOS:
        raise ValueError(
            f"Unexpected premise 2 for rule {application.rule}: {application.premises[2]}"
        )
    eqratio = application.predicate
    if application.premises[3].predicate_type != PredicateType.PARALLEL:
        raise ValueError(
            f"Unexpected premise 3 for rule {application.rule}: {application.premises[3]}"
        )
    para_premise = application.premises[3]

    if application.premises[0].predicate_type != PredicateType.COLLINEAR:
        raise ValueError(
            f"Unexpected premise 0 for rule {application.rule}: {application.premises[0]}"
        )
    coll2 = application.premises[0]
    q1, q2, q3 = coll2.points

    return (
        list(draw_predicate(ax, para_consequence, symbols, theme=theme))
        + list(draw_predicate(ax, eqratio, symbols, theme=theme))
        + list(draw_predicate(ax, para_premise, symbols, theme=theme))
        + [
            draw_segment(
                ax,
                p1.num,
                p2.num,
                line_color=theme.line_color,
                line_width=theme.thick_line_width,
            ),
            draw_segment(
                ax,
                p2.num,
                p3.num,
                line_color=theme.line_color,
                line_width=theme.thick_line_width,
            ),
            draw_segment(
                ax,
                q1.num,
                q2.num,
                line_color=theme.line_color,
                line_width=theme.thick_line_width,
            ),
            draw_segment(
                ax,
                q2.num,
                q3.num,
                line_color=theme.line_color,
                line_width=theme.thick_line_width,
            ),
        ]
    )


def _draw_orthocenter_theorem(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    perps: List[Predicate] = []
    if application.predicate.predicate_type != PredicateType.PERPENDICULAR:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    else:
        perps.append(application.predicate)
    for premise in application.premises:
        if premise.predicate_type != PredicateType.PERPENDICULAR:
            raise ValueError(
                f"Unexpected premise for rule {application.rule}: {premise}"
            )
        else:
            perps.append(premise)
    p1, p2 = application.predicate.line1
    p3, p4 = application.predicate.line2
    predicate_artists: List[Artist] = []
    for perp in perps:
        predicate_artists.extend(
            draw_predicate(ax, perp, symbols, theme=theme)  # symbols are not used here
        )

    return [
        draw_segment(
            ax,
            p1.num,
            p2.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            p1.num,
            p3.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            p1.num,
            p4.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            p2.num,
            p3.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            p2.num,
            p4.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            p3.num,
            p4.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
    ] + predicate_artists


def _draw_incenter_theorem(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    vertices: List[Point] = []

    if len(application.premises) != 3:
        raise ValueError(
            f"Unexpected number of premises for rule {application.rule}: {application.premises}"
        )
    if application.premises[2].predicate_type != PredicateType.N_COLLINEAR:
        raise ValueError(
            f"Unexpected premise 2 for rule {application.rule}: {application.premises[2]}"
        )
    ncoll = application.premises[2]
    for point in ncoll.points:
        if point not in vertices:
            vertices.append(point)
    if application.premises[0].predicate_type != PredicateType.EQUAL_ANGLES:
        raise ValueError(
            f"Unexpected premise 0 for rule {application.rule}: {application.premises[0]}"
        )
    eqangle1 = application.premises[0]
    if application.premises[1].predicate_type != PredicateType.EQUAL_ANGLES:
        raise ValueError(
            f"Unexpected premise 1 for rule {application.rule}: {application.premises[1]}"
        )
    eqangle2 = application.premises[1]

    if application.predicate.predicate_type != PredicateType.EQUAL_ANGLES:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    eqangle3 = application.predicate

    print(application.premises)

    return (
        list(draw_predicate(ax, eqangle1, symbols, theme=theme))
        + list(draw_predicate(ax, eqangle2, symbols, theme=theme))
        + list(draw_predicate(ax, eqangle3, symbols, theme=theme))
        + [
            draw_triangle(
                ax,
                vertices[0].num,
                vertices[1].num,
                vertices[2].num,
                line_color=theme.triangle_color,
                line_width=theme.thick_line_width,
            ),
        ]
    )


def _draw_recognize_center_of_circle(
    ax: Axes,
    application: RuleApplication,
    theme: DrawTheme,
) -> list[Artist]:
    if application.predicate.predicate_type != PredicateType.CONGRUENT:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    if len(application.premises) != 2:
        raise ValueError(
            f"Unexpected number of premises for rule {application.rule}: {application.premises}"
        )
    if application.premises[0].predicate_type != PredicateType.CIRCUMCENTER:
        raise ValueError(
            f"Unexpected premise for rule {application.rule}: {application.premises[0]}"
        )
    circle = application.premises[0]
    cong = application.predicate
    center = circle.center
    point1 = cong.segment1[0]
    point2 = cong.segment1[1]
    point3 = cong.segment2[0]
    point4 = cong.segment2[1]
    radius = point1.num.distance(point2.num)

    return [
        draw_circle(
            ax,
            (center.num.x, center.num.y),
            radius,
            line_color=theme.circle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            point1.num,
            point2.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            point3.num,
            point4.num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        ),
    ]


def _draw_midpoint_consequence(
    ax: Axes,
    application: RuleApplication,
    theme: DrawTheme,
) -> list[Artist]:
    if application.predicate.predicate_type != PredicateType.MIDPOINT:
        raise ValueError(
            f"Unexpected consequence for rule {application.rule}: {application.predicate}"
        )
    midpoint = application.predicate
    p1, p2 = midpoint.segment
    return [
        draw_segment(
            ax,
            p1.num,
            p2.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        )
    ]


def _draw_midpoint(
    ax: Axes,
    application: RuleApplication,
    theme: DrawTheme,
) -> list[Artist]:
    if (
        len(application.premises) < 1
        or application.premises[0].predicate_type != PredicateType.MIDPOINT
    ):
        raise ValueError(
            f"Unexpected premises for rule {application.rule}: {application.premises}"
        )
    midpoint = application.premises[0]
    p1, p2 = midpoint.segment
    return [
        draw_segment(
            ax,
            p1.num,
            p2.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        )
    ]


def _draw_sas_similarity_of_triangles_reverse(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    eqangles: list[Artist] = []
    if application.predicate.predicate_type != PredicateType.SIMTRI_REFLECT:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    a, b, c = application.predicate.triangle1
    p, q, r = application.predicate.triangle2
    for application_premise in application.premises:
        if application_premise.predicate_type == PredicateType.EQUAL_ANGLES:
            artist = draw_predicate(ax, application_premise, symbols, theme=theme)
            eqangles.extend(artist)
    return [
        draw_triangle(
            ax,
            a.num,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_triangle(
            ax,
            p.num,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            a.num,
            b.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            c.num,
            a.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            p.num,
            q.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            r.num,
            p.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
    ] + eqangles


def _draw_sas_similarity_of_triangles_direct(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    eqangles: list[Artist] = []
    if application.predicate.predicate_type != PredicateType.SIMTRI_CLOCK:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    a, b, c = application.predicate.triangle1
    p, q, r = application.predicate.triangle2
    for application_premise in application.premises:
        if application_premise.predicate_type == PredicateType.EQUAL_ANGLES:
            artist = draw_predicate(ax, application_premise, symbols, theme=theme)
            eqangles.extend(artist)
    return [
        draw_triangle(
            ax,
            a.num,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_triangle(
            ax,
            p.num,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            a.num,
            b.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            c.num,
            a.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            p.num,
            q.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            r.num,
            p.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
    ] + eqangles


def _draw_similar_triangles_reverse_properties(
    ax: Axes,
    application: RuleApplication,
    theme: DrawTheme,
) -> list[Artist]:
    simtri: SimtriReflect | None = None
    for premise in application.premises:
        if premise.predicate_type == PredicateType.SIMTRI_REFLECT:
            simtri = premise
            break

    if simtri is None:
        raise ValueError(
            f"Unexpected premises for rule {application.rule}: {application.premises}"
        )
    a, b, c = simtri.triangle1
    p, q, r = simtri.triangle2
    return [
        draw_triangle(
            ax,
            a.num,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_triangle(
            ax,
            p.num,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            a.num,
            b.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            c.num,
            a.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            p.num,
            q.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            r.num,
            p.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
    ]


def _draw_similar_triangles_direct_properties(
    ax: Axes,
    application: RuleApplication,
    theme: DrawTheme,
) -> list[Artist]:
    simtri: SimtriClock | None = None
    for premise in application.premises:
        if premise.predicate_type == PredicateType.SIMTRI_CLOCK:
            simtri = premise
            break

    if simtri is None:
        raise ValueError(
            f"Unexpected premises for rule {application.rule}: {application.premises}"
        )
    a, b, c = simtri.triangle1
    p, q, r = simtri.triangle2
    return [
        draw_triangle(
            ax,
            a.num,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_triangle(
            ax,
            p.num,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            a.num,
            b.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            c.num,
            a.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            p.num,
            q.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            r.num,
            p.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
    ]


def _draw_resolution_of_ratios(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    if application.predicate.predicate_type != PredicateType.EQUAL_RATIOS:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    eqratio_conclusion = application.predicate

    if len(application.premises) != 4:
        raise ValueError(
            f"Unexpected number of premises for rule {application.rule}: {application.premises}"
        )
    if application.premises[2].predicate_type != PredicateType.EQUAL_RATIOS:
        raise ValueError(
            f"Unexpected premise for rule {application.rule}: {application.premises[0]}"
        )
    eqratio_premise = application.premises[2]

    all_segments: set[Artist] = set()
    all_segments.add(
        draw_segment(
            ax,
            eqratio_conclusion.ratio1[0][0].num,
            eqratio_conclusion.ratio1[0][1].num,
            line_color=theme.circle_color,
            line_width=theme.thick_line_width,
        )
    )
    all_segments.add(
        draw_segment(
            ax,
            eqratio_conclusion.ratio1[0][0].num,
            eqratio_conclusion.ratio1[0][1].num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        )
    )
    all_segments.add(
        draw_segment(
            ax,
            eqratio_conclusion.ratio2[0][0].num,
            eqratio_conclusion.ratio2[0][1].num,
            line_color=theme.circle_color,
            line_width=theme.thick_line_width,
        )
    )
    all_segments.add(
        draw_segment(
            ax,
            eqratio_conclusion.ratio2[0][0].num,
            eqratio_conclusion.ratio2[0][1].num,
            line_color=theme.triangle_color,
            line_width=theme.thick_line_width,
        )
    )
    all_segments.update(draw_predicate(ax, eqratio_premise, symbols, theme=theme))

    return list(all_segments)


def _draw_direct_congruence(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    eqangles: list[Artist] = []
    if application.predicate.predicate_type != PredicateType.CONTRI_CLOCK:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    a, b, c = application.predicate.triangle1
    p, q, r = application.predicate.triangle2
    for application_premise in application.premises:
        if application_premise.predicate_type == PredicateType.EQUAL_ANGLES:
            artist = draw_predicate(ax, application_premise, symbols, theme=theme)
            eqangles.extend(artist)
    return [
        draw_triangle(
            ax,
            a.num,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_triangle(
            ax,
            p.num,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            a.num,
            b.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            c.num,
            a.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            p.num,
            q.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            r.num,
            p.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
    ] + eqangles


def _draw_reverse_congruence(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    eqangles: list[Artist] = []
    if application.predicate.predicate_type != PredicateType.CONTRI_REFLECT:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    a, b, c = application.predicate.triangle1
    p, q, r = application.predicate.triangle2
    for application_premise in application.premises:
        if application_premise.predicate_type == PredicateType.EQUAL_ANGLES:
            artist = draw_predicate(ax, application_premise, symbols, theme=theme)
            eqangles.extend(artist)
    return [
        draw_triangle(
            ax,
            a.num,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_triangle(
            ax,
            p.num,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            a.num,
            b.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            b.num,
            c.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            c.num,
            a.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            p.num,
            q.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            q.num,
            r.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
        draw_arrow(
            ax,
            r.num,
            p.num,
            line_color=theme.triangle_color,
            line_width=theme.thin_line_width,
        ),
    ] + eqangles


def _draw_disassembling_a_circle(
    ax: Axes,
    application: RuleApplication,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> list[Artist]:
    circle_premise: Circumcenter | None = None
    for premise in application.premises:
        if premise.predicate_type == PredicateType.CIRCUMCENTER:
            circle_premise = premise
            break

    if (
        circle_premise is None
        or circle_premise.predicate_type != PredicateType.CIRCUMCENTER
    ):
        raise ValueError(
            f"Unexpected premises for rule {application.rule}: {application.premises}"
        )
    radius = circle_premise.center.num.distance(circle_premise.points[0].num)

    if application.predicate.predicate_type != PredicateType.CONGRUENT:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    circle_points: List[Point] = []

    for point in application.predicate.segment1:
        if point != circle_premise.center:
            circle_points.append(point)

    for point in application.predicate.segment2:
        if point != circle_premise.center:
            circle_points.append(point)

    if len(circle_points) != 2:
        raise ValueError(
            f"Unexpected number of points for congruence in rule {application.rule}: {circle_points}"
        )

    return [
        draw_circle(
            ax,
            center=(circle_premise.center.num.x, circle_premise.center.num.y),
            radius=radius,
            line_color=theme.circle_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            circle_premise.center.num,
            circle_points[0].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_segment(
            ax,
            circle_premise.center.num,
            circle_points[1].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
    ]


def _draw_intersection_bisectors(
    ax: Axes,
    application: RuleApplication,
    theme: DrawTheme,
) -> list[Artist]:
    if len(application.premises) != 4:
        raise ValueError(
            f"Unexpected number of premises for rule {application.rule}: {application.premises}"
        )
    if application.predicate.predicate_type != PredicateType.CYCLIC:
        raise ValueError(
            f"Unexpected conclusion for rule {application.rule}: {application.predicate}"
        )
    cyclic = application.predicate
    if application.premises[0].predicate_type != PredicateType.CONGRUENT:
        raise ValueError(
            f"Unexpected premise for rule {application.rule}: {application.premises[0]}"
        )
    cong = application.premises[0]
    if application.premises[1].predicate_type != PredicateType.EQUAL_ANGLES:
        raise ValueError(
            f"Unexpected premise for rule {application.rule}: {application.premises[0]}"
        )
    eqangle = application.premises[1]
    for point1 in cong.segment1:
        for point2 in cong.segment2:
            if point1 == point2:
                intersection = point1
                break

    triangle = [point for point in cyclic.points if point != intersection]

    center = _circumcenter_of_triangle((triangle[0], triangle[1], triangle[2]))
    radius = center.distance(triangle[0].num)
    for point in eqangle.angle1[0]:
        if point not in cong.segment1 and point not in cong.segment2:
            vertex = point
            break

    segment = [point for point in triangle if point != vertex]
    midpoint = _midpoint_of_segment((segment[0], segment[1]))

    return [
        draw_circle(
            ax,
            (center.x, center.y),
            radius,
            line_color=theme.circle_color,
            line_width=theme.thick_line_width,
        ),
        draw_line(
            ax,
            segment[0].num,
            segment[1].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_line(
            ax,
            vertex.num,
            intersection.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_line(
            ax,
            vertex.num,
            segment[0].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_line(
            ax,
            vertex.num,
            segment[1].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_line(
            ax,
            midpoint,
            intersection.num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        ),
        draw_free_perpendicular_symbol(ax, midpoint, intersection.num, theme),
    ]


# Helper functions for geometry calculation


def _one_angle_vertex(angle: tuple[Segment, Segment]) -> Point:
    pair1 = angle[0]
    pair2 = angle[1]
    vertex_set = set(pair1) & set(pair2)
    if not vertex_set:
        raise ValueError("Angle didn't establish a vertex.")
    vertex_name = vertex_set.pop()
    return vertex_name


def _get_eqangle_vertex(predicate: Predicate) -> Point:
    if predicate.predicate_type != PredicateType.EQUAL_ANGLES:
        raise ValueError(
            f"Unexpected predicate type for vertex: {predicate.predicate_type}"
        )
    angle1 = predicate.angle1
    angle2 = predicate.angle2
    vertex1 = _one_angle_vertex(angle1)
    vertex2 = _one_angle_vertex(angle2)

    if vertex1 != vertex2:
        raise ValueError("Equal angle don't have a single vertex.")

    return vertex1


def _midpoint_of_segment(segment: tuple[Point, Point]) -> PointNum:
    if len(segment) != 2:
        raise ValueError("Segment must have exactly 2 points.")

    p1, p2 = (point.num for point in segment)
    midpoint_x = (p1.x + p2.x) / 2
    midpoint_y = (p1.y + p2.y) / 2
    return PointNum(x=midpoint_x, y=midpoint_y)


# FINISH THIS
def _circumcenter_of_triangle(
    triangle: tuple[Point, Point, Point],
) -> PointNum:
    if len(triangle) != 3:
        raise ValueError("Triangle must have exactly 3 points.")

    p1, p2, p3 = (point.num for point in triangle)

    jp1 = JGEXPoint(x=p1.x, y=p1.y)
    jp2 = JGEXPoint(x=p2.x, y=p2.y)
    jp3 = JGEXPoint(x=p3.x, y=p3.y)

    l12 = perpendicular_bisector(jp1, jp2)
    l23 = perpendicular_bisector(jp2, jp3)

    jgex_circumcenter = line_line_intersection(l12, l23)[0]
    circumcenter = PointNum(x=jgex_circumcenter.x, y=jgex_circumcenter.y)
    return circumcenter



import logging
from typing import NamedTuple

import numpy as np
import pytest
from newclid.api import GeometricSolverBuilder
from newclid.deductors.sympy_ar.algebraic_manipulator import SympyARDeductor
from newclid.jgex.problem_builder import JGEXProblemBuilder
from newclid.justifications._index import JustificationType
from newclid.predicates import predicate_from_construction
from newclid.problem import PredicateConstruction, predicate_to_construction
from newclid.proof_justifications import goals_justifications


class ArCase(NamedTuple):
    id: str
    problem: str
    expected_ar_new_predicates: tuple[tuple[str, ...], ...]


#  ∠(BC,BX) = ∠(BY,BC), ∠(AB,BC) = ∠(BX,BY)

AR_CONSEQUENCES_USECASES: list[ArCase] = [
    ArCase(
        id="atable",
        problem="a b c = triangle a b c; x = angle_bisector x a b c; y = angle_mirror y x b c",
        expected_ar_new_predicates=(
            ("eqangle", "a", "b", "b", "x", "b", "c", "b", "y"),
            ("eqangle", "a", "b", "b", "c", "b", "x", "b", "y"),
        ),
    ),
    ArCase(
        id="ieq_triangle",
        problem="a b c = ieq_triangle a b c; x = angle_bisector x a b c; y = angle_mirror y x b a",
        expected_ar_new_predicates=(
            ("perp", "a", "c", "b", "x"),
            ("perp", "b", "c", "b", "y"),
        ),
    ),
    ArCase(
        id="on_aline",
        problem="a b = segment a b; x = free x; y = on_aline y b a x a b",
        expected_ar_new_predicates=(("para", "a", "x", "b", "y"),),
    ),
    ArCase(
        id="aconst",
        problem="a b = segment a b; c = s_angle a b c 30o; d = s_angle c b d 15o",
        expected_ar_new_predicates=(
            ("aconst", "a", "b", "b", "d", "1pi/4"),
            ("aconst", "b", "d", "a", "b", "3pi/4"),
            ("aconst", "b", "c", "a", "b", "5pi/6"),
            ("aconst", "b", "d", "b", "c", "11pi/12"),
        ),
    ),
    ArCase(
        id="rtable",
        problem="a b c = triangle a b c; d = free d; e f g = triangle e f g;"
        " h = eqratio h a b c d e f g; x y z = triangle x y z; w = eqratio w g h e f x y z",
        expected_ar_new_predicates=(
            ("eqratio", "a", "b", "c", "d", "w", "z", "x", "y"),
            ("eqratio", "a", "b", "e", "f", "c", "d", "g", "h"),
            ("eqratio", "a", "b", "w", "z", "c", "d", "x", "y"),
            ("eqratio", "e", "f", "w", "z", "g", "h", "x", "y"),
        ),
    ),
    ArCase(
        id="eqdistance",
        problem="a b = segment a b; c = eqdistance c a a b; d = eqratio d a b a c a c a",
        expected_ar_new_predicates=(
            ("cong", "a", "b", "a", "d"),
            ("cong", "a", "c", "a", "d"),
        ),
    ),
    ArCase(
        id="trisegment",
        problem="a b = segment a b; x y = trisegment x y a b",
        expected_ar_new_predicates=(
            # Congruences
            ("cong", "b", "y", "a", "x"),
            ("cong", "a", "x", "b", "y"),
            ("cong", "a", "y", "b", "x"),
            ("cong", "b", "x", "a", "y"),
            ("rconst", "a", "x", "x", "y", "1/1"),
            ("rconst", "x", "y", "a", "x", "1/1"),
            ("rconst", "b", "y", "x", "y", "1/1"),
            ("rconst", "x", "y", "b", "y", "1/1"),
            ("rconst", "b", "y", "a", "x", "1/1"),
            ("rconst", "a", "x", "b", "y", "1/1"),
            ("rconst", "a", "y", "b", "x", "1/1"),
            ("rconst", "b", "x", "a", "y", "1/1"),
            # Constant ratios
            ("rconst", "a", "b", "x", "y", "3/1"),
            ("rconst", "a", "b", "b", "y", "3/1"),
            ("rconst", "x", "y", "a", "b", "1/3"),
            ("rconst", "a", "b", "a", "x", "3/1"),
            ("rconst", "b", "y", "a", "b", "1/3"),
            ("rconst", "b", "x", "x", "y", "2/1"),
            ("rconst", "b", "x", "a", "x", "2/1"),
            ("rconst", "a", "x", "b", "x", "1/2"),
            ("rconst", "b", "x", "b", "y", "2/1"),
            ("rconst", "b", "x", "a", "b", "2/3"),
            ("rconst", "a", "b", "b", "x", "3/2"),
            ("rconst", "a", "y", "x", "y", "2/1"),
            ("rconst", "a", "y", "a", "x", "2/1"),
            ("rconst", "a", "y", "b", "y", "2/1"),
            ("rconst", "b", "y", "a", "y", "1/2"),
            ("rconst", "a", "y", "a", "b", "2/3"),
            ("rconst", "a", "b", "a", "y", "3/2"),
        ),
    ),
    ArCase(
        id="lconst",
        problem="a = free a; b = lconst b a 4; c = free c; d = eqdistance d c a b",
        expected_ar_new_predicates=(("lconst", "d", "c", "4"),),
    ),
]


# TODO: re-enable this test when AR is ran each level again
@pytest.mark.skip("Skipping for now AR is not ran each level")
@pytest.mark.parametrize(
    argnames="case",
    argvalues=AR_CONSEQUENCES_USECASES,
    ids=[case.id for case in AR_CONSEQUENCES_USECASES],
)
def test_ar_predicates_in_dependency_hypergraph(case: ArCase):
    match_logger = logging.getLogger("newclid.match_theorems")
    match_logger.setLevel(logging.INFO)

    problem = case.problem
    rng = np.random.default_rng(42)
    problem_setup = JGEXProblemBuilder(rng).with_problem_from_txt(problem).build()
    solver = GeometricSolverBuilder(rng).build(problem_setup)
    solver.run()

    expected_goals_constructions = [
        PredicateConstruction.from_tuple(tokens)
        for tokens in case.expected_ar_new_predicates
    ]
    expected_goals = [
        predicate_from_construction(
            construction, points_registry=solver.proof_state.symbols.points
        )
        for construction in expected_goals_constructions
    ]
    expected_goals = [g for g in expected_goals if g is not None]
    goal_predicates = set([g.predicate_type for g in expected_goals])
    goals_constructions = [predicate_to_construction(g) for g in expected_goals]

    try:
        all_predicates_found_by_ar = [
            stmt
            for stmt, dep in solver.proof_state.graph.hyper_graph.items()
            if dep.dependency_type == JustificationType.AR_DEDUCTION
            and stmt.predicate_type in goal_predicates
        ]
        assert set(all_predicates_found_by_ar) == set(expected_goals)

        justifications_for_expected_goals = goals_justifications(
            expected_goals, solver.proof_state
        )
        assert len(justifications_for_expected_goals) > 0
        print(solver.proof(goals_constructions=goals_constructions))
    finally:
        for stmt, justification in solver.proof_state.graph.hyper_graph.items():
            print(stmt, justification)

        ar_deductor = solver.proof_state.deductors[0]
        assert isinstance(ar_deductor, SympyARDeductor)
        print("RTABLE:")
        print(ar_deductor.rtable.inner_table.v2e.keys())  # type: ignore
        print("ATABLE:")
        print(ar_deductor.atable.inner_table.v2e.keys())  # type: ignore


def test_ar_world_hardest_problem_vertex():
    rng = np.random.default_rng(998244353)
    problem_setup = (
        JGEXProblemBuilder(rng)
        .with_problem_from_txt(
            "a b = segment a b; "
            "o = s_angle b a o 70o, s_angle a b o 120o; "
            "c = s_angle o a c 10o, s_angle o b c 160o; "
            "d = on_line d o b, on_line d c a; "
            "e = on_line e o a, on_line e c b; "
            "f = on_pline f d a b, on_line f b c; "
            "g = on_line g f a, on_line g d b "
            "? aconst c a c b 1pi/9"
        )
        .build()
    )
    success = GeometricSolverBuilder(rng).build(problem_setup).run()
    assert success


def test_ar_ratio_hallucination():
    rng = np.random.default_rng(998244353)
    problem_setup = (
        JGEXProblemBuilder(rng)
        .with_problem_from_txt(
            "a b e = triangle12 a b e; c = midpoint c a e ? cong a c a b"
        )
        .build()
    )
    success = GeometricSolverBuilder(rng).build(problem_setup).run()
    assert success



"""Definitions of points that directly satisfy a predicate."""

from newclid.jgex.clause import JGEXClause
from newclid.jgex.constructions._index import JGEXConstructionName
from newclid.jgex.definition import JGEXDefinition, SketchConstruction, v
from newclid.jgex.sketch import (
    sketch_aconst,
    sketch_contri,
    sketch_contrir,
    sketch_eqangle2,
    sketch_eqangle3,
    sketch_eqratio,
    sketch_eqratio6,
    sketch_function_name,
    sketch_l2const,
    sketch_lconst,
    sketch_r2const,
    sketch_rconst,
    sketch_rconst2,
    sketch_s_angle,
    sketch_simtri,
    sketch_simtrir,
)

# Angles

ACONST = JGEXDefinition(
    name=JGEXConstructionName.ACONST,
    args=(v("a"), v("b"), v("c"), v("x"), v("r")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("aconst a b c x r"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_aconst.__name__),
            args=(v("a"), v("b"), v("c"), v("r")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


S_ANGLE = JGEXDefinition(
    name=JGEXConstructionName.S_ANGLE,
    args=(v("a"), v("b"), v("x"), v("y")),
    rely_on_points={v("x"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("aconst a b b x y"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_s_angle.__name__),
            args=(v("a"), v("b"), v("y")),
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"),),
)


EQANGLE2 = JGEXDefinition(
    name=JGEXConstructionName.EQANGLE2,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("eqangle a b a x c x c b"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_eqangle2.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)

EQANGLE3 = JGEXDefinition(
    name=JGEXConstructionName.EQANGLE3,
    args=(v("x"), v("a"), v("b"), v("d"), v("e"), v("f")),
    rely_on_points={v("x"): (v("a"), v("b"), v("d"), v("e"), v("f"))},
    requirements=JGEXClause.from_str("ncoll d e f, diff a b, diff d e, diff e f")[0],
    clauses=JGEXClause.from_str("eqangle x a x b d e d f"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_eqangle3.__name__),
            args=(v("a"), v("b"), v("d"), v("e"), v("f")),
        ),
    ),
    input_points=(v("a"), v("b"), v("d"), v("e"), v("f")),
    output_points=(v("x"),),
)


# Ratios


RCONST = JGEXDefinition(
    name=JGEXConstructionName.RCONST,
    args=(v("a"), v("b"), v("c"), v("x"), v("r")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("rconst a b c x r"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_rconst.__name__),
            args=(v("a"), v("b"), v("c"), v("r")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


RCONST2 = JGEXDefinition(
    name=JGEXConstructionName.RCONST2,
    args=(v("x"), v("a"), v("b"), v("r")),
    rely_on_points={v("x"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("rconst x a x b r"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_rconst2.__name__),
            args=(v("a"), v("b"), v("r")),
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"),),
)

R2CONST = JGEXDefinition(
    name=JGEXConstructionName.R2CONST,
    args=(v("a"), v("b"), v("c"), v("x"), v("r")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("r2const a b c x r"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_r2const.__name__),
            args=(v("a"), v("b"), v("c"), v("r")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)

EQRATIO = JGEXDefinition(
    name=JGEXConstructionName.EQRATIO,
    args=(v("x"), v("a"), v("b"), v("c"), v("d"), v("e"), v("f"), v("g")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"), v("d"), v("e"), v("f"), v("g"))},
    requirements=JGEXClause.from_str("diff a b, diff c d, diff e f")[0],
    clauses=JGEXClause.from_str("eqratio a b c d e f g x"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_eqratio.__name__),
            args=(v("a"), v("b"), v("c"), v("d"), v("e"), v("f"), v("g")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("d"), v("e"), v("f"), v("g")),
    output_points=(v("x"),),
)

EQRATIO6 = JGEXDefinition(
    name=JGEXConstructionName.EQRATIO6,
    args=(v("x"), v("a"), v("c"), v("e"), v("f"), v("g"), v("h")),
    rely_on_points={v("x"): (v("a"), v("c"), v("e"), v("f"), v("g"), v("h"))},
    requirements=JGEXClause.from_str("diff e f, diff g h")[0],
    clauses=JGEXClause.from_str("eqratio a x c x e f g h"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_eqratio6.__name__),
            args=(v("a"), v("c"), v("e"), v("f"), v("g"), v("h")),
        ),
    ),
    input_points=(v("a"), v("c"), v("e"), v("f"), v("g"), v("h")),
    output_points=(v("x"),),
)


# Lengths

LCONST = JGEXDefinition(
    name=JGEXConstructionName.LCONST,
    args=(v("x"), v("a"), v("l")),
    rely_on_points={v("a"): (v("x"),)},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("lconst x a l"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_lconst.__name__), args=(v("a"), v("l"))
        ),
    ),
    input_points=(v("a"),),
    output_points=(v("x"),),
)


L2CONST = JGEXDefinition(
    name=JGEXConstructionName.L2CONST,
    args=(v("x"), v("a"), v("l")),
    rely_on_points={v("a"): (v("x"),)},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("l2const x a l"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_l2const.__name__),
            args=(v("a"), v("l")),
        ),
    ),
    input_points=(v("a"),),
    output_points=(v("x"),),
)

# Similar triangles


SIMTRI = JGEXDefinition(
    name=JGEXConstructionName.SIMTRI,
    args=(v("r"), v("a"), v("b"), v("c"), v("p"), v("q")),
    rely_on_points={v("r"): (v("a"), v("b"), v("c"), v("p"), v("q"))},
    requirements=JGEXClause.from_str("ncoll a b c, diff p q")[0],
    clauses=JGEXClause.from_str("simtri a b c p q r"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_simtri.__name__),
            args=(v("a"), v("b"), v("c"), v("p"), v("q")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("p"), v("q")),
    output_points=(v("r"),),
)


SIMTRIR = JGEXDefinition(
    name=JGEXConstructionName.SIMTRIR,
    args=(v("r"), v("a"), v("b"), v("c"), v("p"), v("q")),
    rely_on_points={v("r"): (v("a"), v("b"), v("c"), v("p"), v("q"))},
    requirements=JGEXClause.from_str("ncoll a b c, diff p q")[0],
    clauses=JGEXClause.from_str("simtrir a b c p q r"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_simtrir.__name__),
            args=(v("a"), v("b"), v("c"), v("p"), v("q")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("p"), v("q")),
    output_points=(v("r"),),
)


CONTRI = JGEXDefinition(
    name=JGEXConstructionName.CONTRI,
    args=(v("q"), v("r"), v("a"), v("b"), v("c"), v("p")),
    rely_on_points={
        v("q"): (v("a"), v("b"), v("c"), v("p")),
        v("r"): (v("a"), v("b"), v("c"), v("p"), v("q")),
    },
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("contri a b c p q r"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_contri.__name__),
            args=(v("a"), v("b"), v("c"), v("p")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("p")),
    output_points=(v("q"), v("r")),
)


CONTRIR = JGEXDefinition(
    name=JGEXConstructionName.CONTRIR,
    args=(v("q"), v("r"), v("a"), v("b"), v("c"), v("p")),
    rely_on_points={
        v("q"): (v("a"), v("b"), v("c"), v("p")),
        v("r"): (v("a"), v("b"), v("c"), v("p"), v("q")),
    },
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("contrir a b c p q r"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_contrir.__name__),
            args=(v("a"), v("b"), v("c"), v("p")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("p")),
    output_points=(v("q"), v("r")),
)

PREDICATE_PRESCRIPTIONS = [
    ACONST,
    S_ANGLE,
    EQANGLE2,
    EQANGLE3,
    RCONST,
    RCONST2,
    R2CONST,
    EQRATIO,
    EQRATIO6,
    LCONST,
    L2CONST,
    SIMTRI,
    SIMTRIR,
    CONTRI,
    CONTRIR,
]



from __future__ import annotations

from typing import Literal, Optional

from newclid.numerical import close_enough
from newclid.numerical.check import same_clock
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.predicates.triangles_similar import two_triangles
from newclid.symbols.points_registry import Triangle


class ContriClock(PredicateInterface):
    """contri A B C P Q R -

    Represent that triangles ABC and PQR are congruent under orientation-preserving
    transformations taking A to P, B to Q and C to R.

    It is equivalent to the three eqangle and eqratio predicates
    on the corresponding angles and sides.
    """

    predicate_type: Literal[PredicateType.CONTRI_CLOCK] = PredicateType.CONTRI_CLOCK
    triangle1: Triangle
    triangle2: Triangle

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> Optional[tuple[PredicateArgument, ...]]:
        return two_triangles(*args)

    def check_numerical(self) -> bool:
        a, b, c = self.triangle1
        p, q, r = self.triangle2
        return (
            close_enough(a.num.distance(b.num), p.num.distance(q.num))
            and close_enough(a.num.distance(c.num), p.num.distance(r.num))
            and close_enough(b.num.distance(c.num), q.num.distance(r.num))
            and same_clock(a.num, b.num, c.num, p.num, q.num, r.num)
        )

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.triangle1 + self.triangle2)

    def __str__(self) -> str:
        a, b, c = self.triangle1
        p, q, r = self.triangle2
        return f"▲{a}{b}{c} ≡ ▲{p}{q}{r}"


class ContriReflect(PredicateInterface):
    """contrir A B C P Q R -

    Represent that triangles ABC and PQR are congruent under orientation-preserving
    transformations taking A to P, B to Q and C to R.

    It is equivalent to the three eqangle and eqratio predicates
    on the corresponding angles and sides.
    """

    predicate_type: Literal[PredicateType.CONTRI_REFLECT] = PredicateType.CONTRI_REFLECT
    triangle1: Triangle
    triangle2: Triangle

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        return two_triangles(*args)

    def check_numerical(self) -> bool:
        a, b, c = self.triangle1
        p, q, r = self.triangle2
        return (
            close_enough(a.num.distance(b.num), p.num.distance(q.num))
            and close_enough(a.num.distance(c.num), p.num.distance(r.num))
            and close_enough(b.num.distance(c.num), q.num.distance(r.num))
            and same_clock(a.num, b.num, c.num, p.num, r.num, q.num)
        )

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.triangle1 + self.triangle2)

    def __str__(self) -> str:
        a, b, c = self.triangle1
        p, q, r = self.triangle2
        return f"▲{a}{b}{c} ≡ ▲{p}{q}{r}"



from dataclasses import dataclass

from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.typing import ColorType
from pydantic import BaseModel

from newclid.animation.artists_in_animation import ArtistInAnimation, PointArtists
from newclid.draw.ar_application import draw_ar_application
from newclid.draw.predicates import draw_predicate
from newclid.draw.rule import draw_rule_application
from newclid.draw.theme import DrawTheme
from newclid.justifications._index import JustificationType
from newclid.justifications.justification import Justification
from newclid.predicate_types import PredicateArgument
from newclid.problem import predicate_points
from newclid.proof_data import ProofData
from newclid.proof_writing import write_proof_sections
from newclid.symbols.symbols_registry import SymbolsRegistry


@dataclass
class JustificationArtists:
    justification: Justification
    artists: list[ArtistInAnimation]
    used_points: list[PointArtists]

    def highlight(
        self,
        aux_point_names: set[PredicateArgument],
        highlight_color: ColorType | None = None,
    ) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for artist in self.artists:
            if artist.highlight(highlight_color):
                updated_artists.add(artist.artist)
        for point in self.used_points:
            color = highlight_color if point.name not in aux_point_names else None
            updated_artists.update(point.highlight(color=color))
        return updated_artists

    def gray_out(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for artist in self.artists:
            if artist.gray_out():
                updated_artists.add(artist.artist)
        return updated_artists

    def hide(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for artist in self.artists:
            if artist.hide():
                updated_artists.add(artist.artist)
        return updated_artists


class JustificationToDraw(BaseModel):
    justification: Justification
    text: str


def init_justifications_to_draw(
    ax: Axes,
    point_name_to_artist: dict[str, PointArtists],
    symbols: SymbolsRegistry,
    proof_data: ProofData,
    theme: DrawTheme,
) -> tuple[
    list[JustificationToDraw],
    dict[Justification, JustificationArtists],
]:
    justifications_to_draw: list[JustificationToDraw] = []
    justification_artists: dict[Justification, JustificationArtists] = {}

    proof_lines_by_sections = write_proof_sections(proof_data)
    for proof_step, step_text in zip(
        proof_data.proof_steps, proof_lines_by_sections.proof_steps
    ):
        justification = proof_step.justification

        used_points_names: set[PredicateArgument]
        used_points: list[PointArtists]
        artists: set[Artist]
        match justification.dependency_type:
            case JustificationType.RULE_APPLICATION:
                used_points_names = set(predicate_points(justification.predicate))
                for premise in justification.premises:
                    used_points_names.update(predicate_points(premise))
                used_points = [
                    point_name_to_artist[pname] for pname in used_points_names
                ]
                artists = set(
                    draw_rule_application(ax, justification, symbols, theme=theme)
                )
            case JustificationType.AR_DEDUCTION:
                if justification.ar_premises is None:
                    raise ValueError(f"No why for dependency: {justification}")

                used_points_names = set(predicate_points(justification.predicate))
                for ar_premise in justification.ar_premises:
                    used_points_names.update(predicate_points(ar_premise.predicate))
                used_points = [
                    point_name_to_artist[pname] for pname in used_points_names
                ]
                artists = draw_ar_application(ax, justification, symbols, theme=theme)
            case _:
                used_points_names = set(predicate_points(justification.predicate))
                used_points = [
                    point_name_to_artist[pname] for pname in used_points_names
                ]
                artists = draw_generic_justification(ax, justification, symbols, theme)

        justification_artists[justification] = JustificationArtists(
            justification=justification,
            artists=[ArtistInAnimation(artist) for artist in artists],
            used_points=used_points,
        )
        justifications_to_draw.append(
            JustificationToDraw(
                justification=justification,
                text=_adapt_text_line(step_text),
            )
        )
    return justifications_to_draw, justification_artists


def _adapt_text_line(text_line: str) -> str:
    text_line = text_line.split("| ")[1]
    if len(text_line) > 110:
        text_line = text_line[:30] + " ... " + text_line[-80:]
    return text_line


def draw_generic_justification(
    ax: Axes,
    justification: Justification,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> set[Artist]:
    artists: set[Artist] = set()
    artists.update(draw_predicate(ax, justification.predicate, symbols, theme=theme))
    return artists



from __future__ import annotations

from typing import Literal

import numpy as np

from newclid.numerical import close_enough
from newclid.numerical.geometries import line_num_from_points
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import NumericalPredicate, PredicateInterface
from newclid.predicates.congruence import Cong
from newclid.symbols.lines_registry import LineSymbol
from newclid.symbols.points_registry import Line
from newclid.symbols.symbols_registry import LineOrCircle, SymbolsRegistry


class Para(PredicateInterface):
    """para A B C D -
    Represent that the line AB is parallel to the line CD.
    """

    predicate_type: Literal[PredicateType.PARALLEL] = PredicateType.PARALLEL
    line1: Line
    line2: Line

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        return Cong.preparse(args)

    def check_numerical(self) -> bool:
        a, b = self.line1
        c, d = self.line2
        line1_angle = (b.num - a.num).angle() % np.pi
        line2_angle = (d.num - c.num).angle() % np.pi
        if (
            not close_enough(line2_angle, line1_angle)
            and not close_enough(line2_angle, line1_angle + np.pi)
            and not close_enough(line2_angle, line1_angle - np.pi)
        ):
            return False
        return True

    def symbols(self, symbols: SymbolsRegistry) -> tuple[LineOrCircle, ...]:
        # Create the lines that are parallel in the symbols graph
        parallels: list[LineSymbol] = []
        a, b = self.line1
        c, d = self.line2
        parallels.append(symbols.lines.line_thru_pair(a, b))
        parallels.append(symbols.lines.line_thru_pair(c, d))
        return tuple(parallels)

    def to_constructive(self, point: str) -> str:
        a, b = [p.name for p in self.line1]
        c, d = [p.name for p in self.line2]
        if point in [c, d]:
            a, b, c, d = c, d, a, b
        if point == b:
            a, b = b, a
        return f"on_pline {a} {b} {c} {d}"

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.line1 + self.line2)

    def __str__(self) -> str:
        a, b = self.line1
        c, d = self.line2
        return f"{a}{b} ∥ {c}{d}"


class NPara(NumericalPredicate):
    """npara A B C D -
    Represent that lines AB and CD are NOT parallel.

    It can only be numerically checked
    (angular coefficient of the equations of the lines are different).
    """

    predicate_type: Literal[PredicateType.N_PARALLEL] = PredicateType.N_PARALLEL

    line1: Line
    line2: Line

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        return Para.preparse(args)

    def check_numerical(self) -> bool:
        a, b = self.line1
        c, d = self.line2
        l1 = line_num_from_points(a.num, b.num)
        l2 = line_num_from_points(c.num, d.num)
        return not l1.is_parallel(l2)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.line1 + self.line2)

    def __str__(self) -> str:
        a, b = self.line1
        c, d = self.line2
        return f"{a}{b} ∦ {c}{d}"



from __future__ import annotations

from typing import Annotated

from pydantic import Field

from newclid.symbols.circles_registry import CirclesRegistry, CircleSymbol
from newclid.symbols.lines_registry import LinesRegistry, LineSymbol
from newclid.symbols.points_registry import PointsRegisty

LineOrCircle = Annotated[LineSymbol | CircleSymbol, Field(discriminator="symbol_type")]


class SymbolsRegistry:
    def __init__(self) -> None:
        self.points = PointsRegisty()
        self.lines = LinesRegistry()
        self.circles = CirclesRegistry()



from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field
from pydantic_core import Url

from newclid.jgex.formulation import JGEXFormulation


class CompetitionIds(Enum):
    IMO = "imo"


class Competition(BaseModel):
    id: CompetitionIds
    name: str
    url: Url


Formulation = Annotated[JGEXFormulation, Field(discriminator="formulation_type")]


class ProblemFormulation(BaseModel):
    name: str
    formulation: Formulation


class Problem(BaseModel):
    """A problem and its multiple formulations."""

    source: Competition
    name: str
    natural_language_formulation: str
    formulations: list[ProblemFormulation]



from __future__ import annotations

from typing import NewType

from pydantic import BaseModel

from newclid.jgex.clause import JGEXClause, JGEXConstruction
from newclid.jgex.constructions._index import JGEXConstructionName
from newclid.predicate_types import PredicateArgument
from newclid.tools import atomize

JGEXVariableName = NewType("JGEXVariableName", str)


class SketchConstruction(BaseModel):
    name: str  # TODO: have an enum / union type instead of generic str for all the possible sketches != jgex defs
    args: tuple[JGEXVariableName, ...]

    @classmethod
    def from_str(cls, s: str) -> SketchConstruction:
        name, *args = atomize(s)
        return cls(name=name, args=tuple(JGEXVariableName(a) for a in args))


class SketchCall(BaseModel):
    name: str
    args: tuple[PredicateArgument, ...]

    @classmethod
    def from_construction(
        cls,
        construction: SketchConstruction,
        mapping: dict[JGEXVariableName, PredicateArgument],
    ) -> SketchCall:
        return cls(
            name=construction.name,
            args=tuple(mapping[arg] for arg in construction.args),
        )


def v(s: str) -> JGEXVariableName:
    return JGEXVariableName(s)


class JGEXDefinition(BaseModel):
    """Definitions of construction predicates."""

    name: JGEXConstructionName
    args: tuple[JGEXVariableName, ...]
    rely_on_points: dict[JGEXVariableName, tuple[JGEXVariableName, ...]]
    requirements: JGEXClause
    clauses: tuple[JGEXClause, ...]
    sketches: tuple[SketchConstruction, ...]
    input_points: tuple[JGEXVariableName, ...] = ()
    output_points: tuple[JGEXVariableName, ...] = ()

    @staticmethod
    def to_dict(
        defs: list[JGEXDefinition],
    ) -> dict[str, JGEXDefinition]:
        return {d.name.value: d for d in defs}


def mapping_from_construction(
    construction: JGEXConstruction,
    construction_definition: JGEXDefinition,
) -> dict[JGEXVariableName, PredicateArgument]:
    if not len(construction.args) == len(construction_definition.args):
        raise ValueError(
            f"Construction {construction} has {len(construction.args)}"
            f" arguments but definition {construction_definition.name}"
            f" has {len(construction_definition.args)}"
        )
    mapping = dict(zip(construction_definition.args, construction.args))
    return mapping


def input_points_of_clause(
    clause: JGEXClause, defs: dict[JGEXConstructionName, JGEXDefinition]
) -> set[PredicateArgument]:
    """Get the input points of a clause."""
    input_points: set[PredicateArgument] = set()
    for construction in clause.constructions:
        cdef = defs[JGEXConstructionName(construction.name)]
        mapping = mapping_from_construction(construction, cdef)
        mapped_inputs = {mapping[p] for p in cdef.input_points}
        input_points.update(mapped_inputs)
    return input_points



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Annotated, Literal

from pydantic import BaseModel, Field

from newclid.agent.agents_interface import DeductiveAgent
from newclid.deductors import ARReason
from newclid.deductors.deductor_interface import ARCoefficient, ARPremise
from newclid.justifications.justification import (
    ARDeduction,
    Justification,
    NumericalCheck,
    Reflexivity,
    RuleApplication,
)
from newclid.justifications.predicates_graph import PredicatesGraph
from newclid.predicates import (
    NUMERICAL_PREDICATES,
    Predicate,
    predicate_from_construction,
)
from newclid.problem import PredicateConstruction, ProblemSetup
from newclid.rule import Rule

if TYPE_CHECKING:
    from newclid.proof_state import ProofState

LOGGER = logging.getLogger(__name__)


class DeductionType(str, Enum):
    RULE = "rule"
    AR = "ar"
    NUM = "num"
    REFLEXIVITY = "refl"


class CachedRuleDeduction(BaseModel):
    deduction_type: Literal[DeductionType.RULE] = DeductionType.RULE
    rule: Rule
    premises: tuple[PredicateConstruction, ...]
    conclusions: tuple[PredicateConstruction, ...]
    point_deps: list[str]

    def __hash__(self) -> int:
        return hash((self.rule, self.premises, self.conclusions))


class ARPremiseConstruction(BaseModel):
    predicate_construction: PredicateConstruction
    coefficient: ARCoefficient

    def __hash__(self) -> int:
        return hash((self.predicate_construction, self.coefficient))


class CachedARDeduction(BaseModel):
    deduction_type: Literal[DeductionType.AR] = DeductionType.AR
    ar_reason: ARReason
    premises: tuple[ARPremiseConstruction, ...]
    conclusions: tuple[PredicateConstruction, ...]
    point_deps: list[str]

    def __hash__(self) -> int:
        return hash((self.ar_reason, self.premises, self.conclusions))


class CachedNumericalCheckDeduction(BaseModel):
    deduction_type: Literal[DeductionType.NUM] = DeductionType.NUM
    conclusions: tuple[PredicateConstruction, ...]

    def __hash__(self) -> int:
        return hash(self.conclusions)


class CachedReflexivityDeduction(BaseModel):
    deduction_type: Literal[DeductionType.REFLEXIVITY] = DeductionType.REFLEXIVITY
    conclusions: tuple[PredicateConstruction, ...]

    def __hash__(self) -> int:
        return hash(self.conclusions)


CachedDeduction = Annotated[
    CachedRuleDeduction
    | CachedARDeduction
    | CachedNumericalCheckDeduction
    | CachedReflexivityDeduction,
    Field(discriminator="deduction_type"),
]


class DeductionProvider(ABC):
    @abstractmethod
    def ordered_deductions_for_problem(
        self, problem: ProblemSetup
    ) -> list[CachedDeduction]: ...

    @property
    @abstractmethod
    def precomputation_input_str(self) -> str: ...


class FollowDeductionsStats(BaseModel):
    """Statistics for the FollowDeductions agent."""

    agent_type: Literal["follow_deductions"] = "follow_deductions"

    n_deductions_stored: int
    """The number of deductions stored."""

    n_deductions_followed: int
    """The list of deductions followed."""


class FollowDeductions(DeductiveAgent):
    def __init__(self, deductions_provider: DeductionProvider) -> None:
        self._deductions_provider = deductions_provider
        self._deductions: list[CachedDeduction] = []
        self.premises_of_deduction: dict[CachedDeduction, list[Predicate]] = {}
        self.deps_of_deduction: dict[CachedDeduction, list[Justification]] = {}
        self._has_gathered_deductions = False
        self._stats = FollowDeductionsStats(
            n_deductions_stored=0, n_deductions_followed=0
        )

    def get_stats(self) -> FollowDeductionsStats:
        return self._stats.model_copy()

    def step(self, proof: ProofState, rules: list[Rule]) -> bool:
        if not self._has_gathered_deductions:
            self._gather_deductions(proof)

        if not self._deductions:
            # No more deductions to follow, we're done.
            return False

        next_deduction = self._deductions.pop(0)
        precomputation_input_str = self._deductions_provider.precomputation_input_str
        premises_predicates = self.premises_of_deduction[next_deduction]
        _check_premises_of_deduction(
            next_deduction,
            precomputation_input_str=precomputation_input_str,
            premises_predicates=premises_predicates,
            pred_graph=proof.graph,
        )
        for new_dep in self.deps_of_deduction[next_deduction]:
            LOGGER.debug(f"Proved by following deduction: {new_dep}")
            _success = proof.apply(new_dep)
            self._stats.n_deductions_followed += 1

        return True

    def _gather_deductions(self, proof_state: ProofState) -> None:
        self._deductions = self._deductions_provider.ordered_deductions_for_problem(
            proof_state.problem
        ).copy()
        self._stats.n_deductions_stored = len(self._deductions)
        for deduction in self._deductions:
            self.premises_of_deduction[deduction] = _validate_premises_of_deduction(
                deduction,
                precomputation_input_str=self._deductions_provider.precomputation_input_str,
                proof_state=proof_state,
            )
            self.deps_of_deduction[deduction] = _deps_from_conclusions_of_deduction(
                deduction,
                precomputation_input_str=self._deductions_provider.precomputation_input_str,
                premises_predicates=self.premises_of_deduction[deduction],
                proof_state=proof_state,
            )
        self._has_gathered_deductions = True

    def reset(self) -> None:
        self._deductions = []
        self.premises_of_deduction = {}
        self.deps_of_deduction = {}
        self._has_gathered_deductions = False
        self._stats = FollowDeductionsStats(
            n_deductions_stored=0, n_deductions_followed=0
        )


def _validate_premises_of_deduction(
    deduction: CachedDeduction,
    precomputation_input_str: str,
    proof_state: ProofState,
) -> list[Predicate]:
    match deduction.deduction_type:
        case DeductionType.NUM | DeductionType.REFLEXIVITY:
            return []
        case DeductionType.RULE:
            premises = deduction.premises
        case DeductionType.AR:
            premises = tuple(
                premise.predicate_construction for premise in deduction.premises
            )

    premises_predicates: list[Predicate] = []
    for premise in premises:
        premise_predicate = predicate_from_construction(
            premise, points_registry=proof_state.symbols.points
        )
        if premise_predicate is None:
            error_msg = (
                f"Premise '{premise}' of deduction {deduction} could not be built."
            )
            LOGGER.error(f"{error_msg}.\nSetup:\n{precomputation_input_str}")
            raise ValueError(error_msg)

        if (
            premise_predicate.predicate_type in NUMERICAL_PREDICATES
            and not proof_state.check_numerical(premise_predicate)
        ):
            error_msg = (
                f"Premise '{premise}' is numericaly false within deduction {deduction}."
            )
            LOGGER.error(f"{error_msg}.\nSetup:\n{precomputation_input_str}")
            raise ValueError(error_msg)

        premises_predicates.append(premise_predicate)

    return premises_predicates


def _deps_from_conclusions_of_deduction(
    deduction: CachedDeduction,
    proof_state: ProofState,
    precomputation_input_str: str,
    premises_predicates: list[Predicate],
) -> list[Justification]:
    new_deps: list[Justification] = []
    for conclusion in deduction.conclusions:
        new_predicate = predicate_from_construction(
            conclusion, points_registry=proof_state.symbols.points
        )
        if new_predicate is None:
            error_msg = f"Conclusion {conclusion} could not be built."
            LOGGER.error(f"{error_msg}.\nSetup:\n{precomputation_input_str}")
            raise ValueError(error_msg)

        if not proof_state.check_numerical(new_predicate):
            error_msg = f"Conclusion {new_predicate} is numerical false."
            LOGGER.error(f"{error_msg}.\nSetup:\n{precomputation_input_str}")
            raise ValueError(error_msg)

        if deduction.deduction_type is DeductionType.RULE:
            _check_rule_conclusions(deduction)

        new_dep: Justification
        match deduction.deduction_type:
            case DeductionType.RULE:
                new_dep = RuleApplication(
                    predicate=new_predicate,
                    rule=deduction.rule,
                    premises=tuple(premises_predicates),
                )
            case DeductionType.AR:
                new_dep = ARDeduction(
                    predicate=new_predicate,
                    ar_reason=deduction.ar_reason,
                    ar_premises=tuple(
                        ARPremise(
                            predicate=premise,
                            coefficient=premise_construction.coefficient,
                        )
                        for premise_construction, premise in zip(
                            deduction.premises, premises_predicates
                        )
                    ),
                )
            case DeductionType.NUM:
                new_dep = NumericalCheck(predicate=new_predicate)
            case DeductionType.REFLEXIVITY:
                new_dep = Reflexivity(predicate=new_predicate)

        new_deps.append(new_dep)
    return new_deps


class DoubleCheckError(Exception):
    """A deduction given is not valid to be followed."""


def _check_premises_of_deduction(
    deduction: CachedDeduction,
    precomputation_input_str: str,
    premises_predicates: list[Predicate],
    pred_graph: PredicatesGraph,
) -> None:
    for premise_predicate in premises_predicates:
        premise_justification = pred_graph.hyper_graph.get(premise_predicate)
        if premise_justification is None:
            error_msg = (
                f"Premise '{premise_predicate}' of deduction {deduction} not found"
                " in proof state while following given deductions."
            )
            LOGGER.error(f"{error_msg}.\nSetup:\n{precomputation_input_str}")
            raise DoubleCheckError(error_msg)


def _check_rule_conclusions(deduction: CachedRuleDeduction) -> None:
    """Double check that the deduction is a valid application of the rule."""
    pass



from enum import Enum
from typing import Annotated, Literal
from xml.etree.ElementTree import Element

from pydantic import BaseModel, Field

from newclid.numerical.geometries import PointNum
from newclid.predicate_types import PredicateArgument
from newclid.symbols.points_registry import Point


class GGBElementType(Enum):
    POINT = "point"
    LINE = "line"
    CONIC = "conic"
    SEGMENT = "segment"


class GGBPoint(BaseModel):
    type: Literal[GGBElementType.POINT] = GGBElementType.POINT
    label: str
    x: float
    y: float
    z: float

    @classmethod
    def from_xml_element(cls, element: Element) -> "GGBPoint":
        (coords,) = element.iter("coords")
        x, y, z = map(
            float, (coords.attrib["x"], coords.attrib["y"], coords.attrib["z"])
        )
        return cls(label=element.attrib["label"], x=x, y=y, z=z)

    def to_newclid(self) -> Point:
        return Point(
            name=PredicateArgument(self.label),
            num=PointNum(x=self.x / self.z, y=self.y / self.z),
        )


class GGBLine(BaseModel):
    type: Literal[GGBElementType.LINE] = GGBElementType.LINE
    line_name: str
    points: list[str] = []

    @classmethod
    def from_xml_element(cls, element: Element) -> "GGBLine":
        return cls(line_name=element.attrib["label"])


class GGBConic(BaseModel):
    type: Literal[GGBElementType.CONIC] = GGBElementType.CONIC
    conic_name: str
    points: list[str] = []
    center: str | None = None
    radius: str | tuple[str, str] | None = None

    @classmethod
    def from_xml_element(cls, element: Element) -> "GGBConic":
        return cls(conic_name=element.attrib["label"])


GGBElement = Annotated[GGBLine | GGBPoint | GGBConic, Field(discriminator="type")]


def read_elements(
    root: Element,
) -> tuple[dict[str, GGBPoint], dict[str, GGBLine], dict[str, GGBConic]]:
    points: dict[str, GGBPoint] = {}
    lines: dict[str, GGBLine] = {}
    conics: dict[str, GGBConic] = {}
    for e in root.iter("element"):
        match GGBElementType(e.attrib["type"]):
            case GGBElementType.POINT:
                show = e.find("show")
                if show is not None and show.attrib["object"] == "true":
                    point = GGBPoint.from_xml_element(e)
                    points[point.label] = point
            case GGBElementType.LINE:
                line = GGBLine.from_xml_element(e)
                lines[line.line_name] = line
            case GGBElementType.CONIC:
                conic = GGBConic.from_xml_element(e)
                conics[conic.conic_name] = conic
            case GGBElementType.SEGMENT:
                pass  # TODO: handle segments or do they have line already?

    return points, lines, conics



"""Heuristics that allows to find new potential auxiliary points in a problem."""

from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic, HeuristicSetup

__all__ = [
    "HeuristicName",
    "Heuristic",
    "HeuristicSetup",
]



"""Subpackage for animation of Newclid proofs."""

from __future__ import annotations

from pathlib import Path
from tempfile import TemporaryDirectory

from matplotlib.animation import FuncAnimation, HTMLWriter

from newclid.animation.proof_animation import ProofAnimation

__all__ = [
    "ProofAnimation",
    "html_animation",
]


def html_animation(ani: FuncAnimation) -> str:
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir, "temp.html")
        writer = HTMLWriter(embed_frames=True, fps=2, embed_limit=100_000_000)
        writer.frame_format = "jpeg"
        ani.save(str(path), writer=writer)
        html_representation = path.read_text()
    return html_representation



from newclid.deductors.deductor_interface import Deductor
from newclid.justifications.justification import (
    DirectConsequence,
    Justification,
    NumericalCheck,
)
from newclid.predicates import Predicate
from newclid.predicates._index import PredicateType
from newclid.predicates.congruence import Cong
from newclid.predicates.constant_angle import ACompute, aconst_from_acompute
from newclid.predicates.constant_length import LCompute, lconst_from_lcompute
from newclid.predicates.constant_ratio import RCompute, rconst_from_rcompute
from newclid.symbols.symbols_registry import SymbolsRegistry


def justify_predicate(
    predicate: Predicate,
    deductors: list[Deductor],
    symbols: SymbolsRegistry,
) -> Justification | None:
    match predicate.predicate_type:
        # From symbols
        case PredicateType.COLLINEAR:
            return symbols.lines.why_colllinear(predicate)
        case PredicateType.CYCLIC:
            return symbols.circles.why_cyclic(predicate)

        # By definition
        case PredicateType.CIRCUMCENTER:
            o = predicate.center
            p0 = predicate.points[0]
            why = tuple(
                Cong(segment1=(o, p0), segment2=(o, pi)) for pi in predicate.points[1:]
            )
            return DirectConsequence(predicate=predicate, premises=why)

        case (
            PredicateType.A_COMPUTE | PredicateType.L_COMPUTE | PredicateType.R_COMPUTE
        ):
            constant_predicate = _constant_predicate_from_compute(predicate)
            constant_predicate_justification = justify_predicate(
                constant_predicate, deductors=deductors, symbols=symbols
            )
            if constant_predicate_justification is None:
                # We cannot compute the constant predicate
                return None
            return DirectConsequence(
                predicate=predicate, premises=(constant_predicate,)
            )

        # Numerical check
        case (
            PredicateType.DIFFERENT
            | PredicateType.N_COLLINEAR
            | PredicateType.N_PERPENDICULAR
            | PredicateType.N_PARALLEL
            | PredicateType.SAME_SIDE
            | PredicateType.N_SAME_SIDE
            | PredicateType.SAME_CLOCK
            | PredicateType.OBTUSE_ANGLE
        ):
            return NumericalCheck(predicate=predicate)

        case _:
            return None


def _constant_predicate_from_compute(
    predicate: ACompute | LCompute | RCompute,
) -> Predicate:
    match predicate.predicate_type:
        case PredicateType.A_COMPUTE:
            return aconst_from_acompute(predicate)
        case PredicateType.L_COMPUTE:
            return lconst_from_lcompute(predicate)
        case PredicateType.R_COMPUTE:
            return rconst_from_rcompute(predicate)



from __future__ import annotations

from typing import Literal

from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Point, Segment


class MidPoint(PredicateInterface):
    """midp M A B -
    Represent that M is the midpoint of the segment AB.

    Can be equivalent to coll M A B and cong A M B M."""

    predicate_type: Literal[PredicateType.MIDPOINT] = PredicateType.MIDPOINT
    segment: Segment
    midpoint: Point

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        if len(set(args)) != 3:
            return None
        m, a, b = args
        a, b = sorted((a, b))
        return (m, a, b)

    def check_numerical(self) -> bool:
        a, b = self.segment
        return self.midpoint.num.close_enough((a.num + b.num) / 2)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        a, b = self.segment
        return (
            PredicateArgument(self.midpoint.name),
            PredicateArgument(a.name),
            PredicateArgument(b.name),
        )

    def __str__(self) -> str:
        a, b = self.segment
        return f"{self.midpoint} is the midpoint of {a}{b}"



"""Action / Feedback interface

Make all interactions explicit between DeductiveAgent and the Proof state to allow
for independent developpement of different kinds of DeductiveAgent.

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from newclid.agent import AgentStats
    from newclid.proof_state import ProofState
    from newclid.rule import Rule


class DeductiveAgent(ABC):
    """Common interface for deductive agents"""

    @abstractmethod
    def step(self, proof: ProofState, rules: list[Rule]) -> bool:
        """Perform a single reasoning step on the given proof with given rules, and return if the agent is exausted.

        Returns:
            True if the agent is considered exausted, False otherwise.
        """

    def get_stats(self) -> AgentStats | None:
        """Get the statistics of the agent."""
        return None



from __future__ import annotations

from typing import Literal

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Segment


class Cong(PredicateInterface):
    """cong A B C D -
    Represent that segments AB and CD are congruent."""

    predicate_type: Literal[PredicateType.CONGRUENT] = PredicateType.CONGRUENT
    segment1: Segment
    segment2: Segment

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        segs: list[tuple[PredicateArgument, PredicateArgument]] = []
        if len(args) % 2 != 0:
            return None
        for a, b in zip(args[::2], args[1::2]):
            if a > b:
                a, b = b, a
            if a == b:
                return None
            segs.append((a, b))
        segs.sort()
        segs = list(set(segs))
        points: list[PredicateArgument] = []
        if len(segs) == 0:
            return None

        elif len(segs) == 1:
            a, b = segs[0]
            points.append(a)
            points.append(b)
            points.append(a)
            points.append(b)

        else:
            for a, b in segs:
                points.append(a)
                points.append(b)
        return tuple(points)

    def check_numerical(self) -> bool:
        a, b = self.segment1
        l1 = a.num.distance2(b.num)
        c, d = self.segment2
        l2 = c.num.distance2(d.num)
        return close_enough(l1, l2)

    def to_constructive(self, point: str) -> str:
        a, b, c, d = self.segment1 + self.segment2
        if point in [c.name, d.name]:
            a, b, c, d = c, d, a, b
        if point == b.name:
            a, b = b, a
        if point == d.name:
            c, d = d, c
        if a.name == c.name and a.name == point:
            return f"on_bline {a.name} {b.name} {d.name}"
        if b.name in [c.name, d.name]:
            if b.name == d.name:
                c, d = d, c
            return f"on_circle {a.name} {b.name} {d.name}"
        return f"eqdistance {a.name} {b.name} {c.name} {d.name}"

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.segment1 + self.segment2)

    def __str__(self) -> str:
        a, b = self.segment1
        c, d = self.segment2
        return f"{a}{b} = {c}{d}"



from __future__ import annotations

from fractions import Fraction
from typing import TYPE_CHECKING, Literal

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.predicates.constant_angle import ACompute
from newclid.symbols.points_registry import Ratio
from newclid.tools import fraction_to_ratio, get_quotient, str_to_fraction

if TYPE_CHECKING:
    from newclid.proof_state import ProofState


class ConstantRatio(PredicateInterface):
    """rconst A B C D r -
    Represent that AB / CD = r

    r should be given with numerator and denominator separated by '/', as in 2/3.
    """

    predicate_type: Literal[PredicateType.CONSTANT_RATIO] = PredicateType.CONSTANT_RATIO
    ratio: Ratio
    value: Fraction

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, c, d, r = args
        if a == b or c == d:
            return None
        f = str_to_fraction(r)
        a, b = sorted((a, b))
        c, d = sorted((c, d))
        if (a, b) > (c, d):
            a, b, c, d = c, d, a, b
            f = 1 / f
        return (a, b, c, d, PredicateArgument(fraction_to_ratio(f)))

    def check_numerical(self) -> bool:
        (a, b), (c, d) = self.ratio
        ratio = a.num.distance(b.num) / c.num.distance(d.num)
        return close_enough(ratio, float(self.value))

    def __str__(self) -> str:
        (a, b), (c, d) = self.ratio
        return f"{a}{b}:{c}{d} = {fraction_to_ratio(self.value)}"

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        (a, b), (c, d) = self.ratio
        return (
            PredicateArgument(a.name),
            PredicateArgument(b.name),
            PredicateArgument(c.name),
            PredicateArgument(d.name),
            PredicateArgument(fraction_to_ratio(self.value)),
        )


class RCompute(PredicateInterface):
    """rcompute A B C D"""

    predicate_type: Literal[PredicateType.R_COMPUTE] = PredicateType.R_COMPUTE

    ratio: Ratio

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        return ACompute.preparse(args)

    def check_numerical(self) -> bool:
        return True  # Can always compute ratio

    def check(self, proof_state: ProofState) -> bool | None:
        rconst = rconst_from_rcompute(self)
        return proof_state.check(rconst)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(
            PredicateArgument(p.name) for segment in self.ratio for p in segment
        )

    def __str__(self) -> str:
        return str(rconst_from_rcompute(self))


def rconst_from_rcompute(rcompute: RCompute) -> ConstantRatio:
    (a, b), (c, d) = rcompute.ratio
    ratio = get_quotient(a.num.distance(b.num) / c.num.distance(d.num))
    return ConstantRatio(ratio=rcompute.ratio, value=ratio)



from enum import Enum

"""
Definitions are used in the translation of a problem to a JGEX string to create the coordinates of the points and add the hypothesis predicates to the problem. They correspond to geometric constructions.

Each definition is an instance of the JGEXDefinition class, and it receives the following arguments:

- `name`: The name of the construction, which is the syntax for using it in the JGEX string.
- `args`: The points that the construction requires as arguments, in the order to be declared in the JGEX string.
- `rely_on_points`: Registers the logical dependency between one or many output points and other arguments of the construction.
- `requirements`: A list of numerical predicates that must be satisfied by the input points for the construction to be attempted.
- `clauses`: The list of predicate hypothesis that will be added to the problem by the definition.
- `sketches`: The name of the sketch function called by the definition with its arguments, which creates coordinates for the points given by the definition.
- `input_points`: The previously existing points that the construction requires as input.
- `output_points`: The points that the construction will create as output.

For example in:

ON_CIRCLE = JGEXDefinition(
    name=JGEXConstructionName.ON_CIRCLE,
    args=(v("x"), v("o"), v("a")),
    rely_on_points={v("x"): (v("o"), v("a"))},
    requirements=JGEXClause.from_str("diff o a")[0],
    clauses=JGEXClause.from_str("cong o x o a"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_circle.__name__),
            args=(v("o"), v("o"), v("a")),
        ),
    ),
    input_points=(v("o"), v("a")),
    output_points=(v("x"),),
)

The name, to be found in the JGEXConstructionName enum below, is on_circle. The args are x, o, a (typed as variables).

As x is the output point, and o, a, are input points, the definition would occur in a JGEX string as:

    x = on_circle x o a

o, a should be defined beforehand. The requirement says the engine will check if o, a are distinct points before attempting the construction. If succeeded, the clauses say the hypothesis

    cong o x o a

will be added to the problem.

To get the coordinates of x, the sketch function sketch_circle will be called with the arguments o, o, a.


This file contains the JGEXConstructionName class, which lists all the definitions that can be currently used, separated by categories. The left hand side gives the name of the instance of the definition in the code, while the right hand side gives the syntax to be used in the JGEX string.
"""


class JGEXConstructionName(str, Enum):
    # Free constructions
    FREE = "free"
    SEGMENT = "segment"
    BETWEEN = "between"
    TRIANGLE = "triangle"
    R_TRIANGLE = "r_triangle"
    ACUTE_TRIANGLE = "acute_triangle"
    ISO_TRIANGLE = "iso_triangle"
    ISO_TRIANGLE0 = "iso_triangle0"
    IEQ_TRIANGLE = "ieq_triangle"
    RISOS = "risos"
    TRIANGLE12 = "triangle12"
    QUADRANGLE = "quadrangle"
    RECTANGLE = "rectangle"
    ISQUARE = "isquare"
    TRAPEZOID = "trapezoid"
    R_TRAPEZOID = "r_trapezoid"
    ISO_TRAPEZOID = "iso_trapezoid"
    EQ_QUADRANGLE = "eq_quadrangle"
    EQDIA_QUADRANGLE = "eqdia_quadrangle"
    PENTAGON = "pentagon"

    # Point on object constructions
    ON_LINE = "on_line"
    ON_PARA_LINE = "on_pline"
    ON_PARA_LINE0 = "on_pline0"
    ON_PERP_LINE = "on_tline"
    ON_ALINE = "on_aline"
    ON_ALINE0 = "on_aline0"
    ON_BLINE = "on_bline"
    ON_CIRCLE = "on_circle"
    ON_CIRCUM = "on_circum"
    ON_DIA = "on_dia"

    # Intersection constructions
    INTERSECTION_LINE_LINE = "intersection_ll"
    INTERSECTION_LINE_PARA = "intersection_lp"
    INTERSECTION_PARA_PARA = "intersection_pp"
    INTERSECTION_LINE_PERP = "intersection_lt"
    INTERSECTION_PERP_PERP = "intersection_tt"
    INTERSECTION_LINE_CIRCLE = "intersection_lc"
    INTERSECTION_CIRCLE_CIRCLE = "intersection_cc"

    # Tangent constructions
    TANGENT = "tangent"
    LINE_CIRCLE_TANGENT = "lc_tangent"
    CIRCLE_CIRCLE_TANGENT = "cc_tangent"
    CIRCLE_CIRCLE_ITANGENT = "cc_itangent"

    # Relative to other constructions
    ANGLE_BISECTOR = "angle_bisector"
    ANGLE_MIRROR = "angle_mirror"
    EQDISTANCE = "eqdistance"
    FOOT = "foot"
    INCENTER = "incenter"
    INCENTER2 = "incenter2"
    EXCENTER = "excenter"
    EXCENTER2 = "excenter2"
    MIDPOINT = "midpoint"
    MIRROR = "mirror"
    ORTHOCENTER = "orthocenter"
    REFLECT = "reflect"
    SHIFT = "shift"
    TRISECT = "trisect"
    TRISEGMENT = "trisegment"
    BETWEEN_BOUND = "between_bound"

    # Complete figure constructions
    ISO_TRIANGLE_VERTEX = "iso_triangle_vertex"
    ISO_TRIANGLE_VERTEX_ANGLE = "iso_triangle_vertex_angle"
    EQ_TRIANGLE = "eq_triangle"
    ISO_TRAPEZOID2 = "iso_trapezoid2"
    PARALLELOGRAM = "parallelogram"
    SQUARE = "square"
    CIRCLE = "circle"
    CIRCUMCENTER = "circumcenter"
    PSQUARE = "psquare"
    NSQUARE = "nsquare"
    CENTROID = "centroid"

    # Predicate prescriptions
    ACONST = "aconst"
    S_ANGLE = "s_angle"
    EQANGLE2 = "eqangle2"
    EQANGLE3 = "eqangle3"
    RCONST = "rconst"
    RCONST2 = "rconst2"
    R2CONST = "r2const"
    EQRATIO = "eqratio"
    EQRATIO6 = "eqratio6"
    LCONST = "lconst"
    L2CONST = "l2const"
    SIMTRI = "simtri"
    SIMTRIR = "simtrir"
    CONTRI = "contri"
    CONTRIR = "contrir"

    # Problem specific constructions
    TWOL1C = "2l1c"
    E5128 = "e5128"
    THREEPEQ = "3peq"
    TEST_R20 = "test_r20"
    TEST_R25 = "test_r25"
    NINEPOINTS = "ninepoints"



from __future__ import annotations

from typing import Iterator, Literal

from pydantic import BaseModel

from newclid.justifications._index import JustificationError, JustificationType
from newclid.numerical.geometries import LineNum, line_num_from_points
from newclid.predicates.collinearity import Coll
from newclid.symbols._index import SymbolType
from newclid.symbols.merging_symbols import (
    SymbolMergeHistory,
    merge_symbols,
    representent_of,
)
from newclid.symbols.points_registry import Point


class LineSymbol(BaseModel):
    """
    The LineSymbol class represents a line identified by a coll predicate in the symbolic registry.
    Its attributes are:
    - points: the set of Point objects known to belong to the line
    - num: the LineNum object that contains the numerical parameters of the line
    - justification: a Coll predicate that justifies the existence of this line in the symbols registry
    - symbol_type: a constant indicating that this is a line symbol (SymbolType.LINE).
    """

    points: set[Point]
    num: LineNum
    justification: Coll | None  # TODO: Should not be able to be None
    symbol_type: Literal[SymbolType.LINE] = SymbolType.LINE

    def __str__(self) -> str:
        sorted_point_names = sorted(p.name for p in self.points)
        return "Line(" + "-".join(sorted_point_names) + ")"

    def __repr__(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        return hash(str(self))


class LinesRegistry:
    def __init__(self) -> None:
        self._lines: list[LineSymbol] = []
        self._merge_history: dict[LineSymbol, SymbolMergeHistory[LineSymbol]] = {}

    def create_line_thru_points(
        self, points: set[Point], because: Coll | None = None
    ) -> LineSymbol:
        if len(points) < 2:
            raise ValueError("Line needs at least two points")
        p1, p2 = list(points)[:2]
        line_num = line_num_from_points(p1.num, p2.num)
        line = LineSymbol(justification=because, points=points, num=line_num)
        self._lines.append(line)
        self._merge_history[line] = SymbolMergeHistory(symbol=line)
        return line

    def line_thru_pair(self, p1: Point, p2: Point) -> LineSymbol:
        for line in self._lines:
            if {p1, p2} <= line.points:
                for line1 in self._merge_history[line].fellows:
                    if {p1, p2} == line1.points:
                        return line1
                other_line = self.create_line_thru_points({p1, p2}, because=None)
                assert line.justification
                merge_symbols(
                    self._merge_history[line],
                    [self._merge_history[other_line]],
                    self._lines,
                )
                return other_line
        return self.create_line_thru_points({p1, p2}, because=None)

    def representent(self, line: LineSymbol) -> LineSymbol:
        return representent_of(self._merge_history[line]).symbol

    def line_containing(self, pnames: set[Point]) -> LineSymbol | None:
        for line in self._lines:
            if pnames <= line.points:
                return line
        return None

    def check_collinear(
        self,
        points: tuple[Point, ...],
    ) -> bool:
        s = set(points)
        for line in self._lines:
            if s <= line.points:
                return True
        return False

    def make_collinear(
        self, points: tuple[Point, ...], justification: Coll
    ) -> tuple[LineSymbol, list[LineSymbol]]:
        points_set = set(points)
        merged: list[LineSymbol] = []
        for other_line in self._lines:
            if points_set <= other_line.points:
                # Already belong to the an existing line
                return other_line, []
            if len(points_set & other_line.points) >= 2:
                merged.append(other_line)
                points_set.update(other_line.points)

        line = self.create_line_thru_points(points_set, because=justification)
        merge_symbols(
            self._merge_history[line],
            [self._merge_history[line] for line in merged],
            self._lines,
        )
        return line, merged

    def why_colllinear(self, coll: Coll) -> LineMerge:
        points: tuple[Point, ...] = coll.points
        s = set(points)
        for line in self._lines:
            if not s <= line.points:
                continue
            line_target = line
            for _target in self._merge_history[line].fellows:
                if s <= _target.points and len(_target.points) < len(
                    line_target.points
                ):
                    line_target = _target
            if not line_target.justification:
                raise ValueError(
                    f"Line {line_target} was merged in the symbols graph without justification."
                )
            return LineMerge(
                predicate=coll,
                line=line_target,
                direct_justification=line_target.justification,
            )
        raise JustificationError("Could not justify why the points would be collinear")

    def get_all_points_on_line_and_equivalent_lines(
        self, line: LineSymbol
    ) -> set[Point]:
        points: set[Point] = set()
        for equivalent_line in self._merge_history[line].fellows:
            for pt in equivalent_line.points:
                points.add(pt)
        return points

    def __iter__(self) -> Iterator[LineSymbol]:
        return iter(self._lines)


class LineMerge(BaseModel):
    predicate: Coll
    """The predicate that is justified by the symbols graph merge."""
    line: LineSymbol
    """The line symbol that was merged."""
    direct_justification: Coll
    """The previous predicate that directly justifies the merge that has the predicate as a consequence."""

    dependency_type: Literal[JustificationType.LINE_MERGE] = (
        JustificationType.LINE_MERGE
    )



import logging

import numpy

from newclid.api import GeometricSolver, GeometricSolverBuilder
from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic, HeuristicSetup
from newclid.heuristics.angle_vertices import (
    AngleVerticesHeuristicConfig,
)
from newclid.heuristics.centers_of_cyclic import (
    CentersOfCyclicHeuristicConfig,
)
from newclid.heuristics.geometric_objects import (
    read_geometric_objects_from_predicate_constructions,
)
from newclid.heuristics.heuristic_from_config import (
    HeuristicConfig,
    heuristic_from_config,
)
from newclid.heuristics.line_intersections import LineIntersectionsHeuristicConfig
from newclid.jgex.constructions import ALL_JGEX_CONSTRUCTIONS
from newclid.jgex.definition import JGEXDefinition
from newclid.jgex.errors import JGEXConstructionError
from newclid.jgex.formulation import JGEXFormulation
from newclid.jgex.problem_builder import JGEXProblemBuilder
from newclid.jgex.to_newclid import JGEXClauseConsequences, add_clause_to_problem
from newclid.predicate_types import PredicateArgument
from newclid.problem import ProblemSetup, predicate_to_construction

defs = JGEXDefinition.to_dict(ALL_JGEX_CONSTRUCTIONS)
LOGGER = logging.getLogger(__name__)
rng = numpy.random.default_rng(42)


def apply_complete_the_picture_heuristics(
    problem_setup: ProblemSetup,
    jgex_problem: JGEXFormulation,
    rng: numpy.random.Generator,
    max_new_points: int | None = None,
) -> tuple[
    ProblemSetup, JGEXFormulation, dict[HeuristicName, list[JGEXClauseConsequences]]
]:
    """Apply the "complete the picture" heuristics to the given JGEX problem."""
    complete_the_picture_heuristics_configs: list[HeuristicConfig] = [
        CentersOfCyclicHeuristicConfig(),
        LineIntersectionsHeuristicConfig(),
        AngleVerticesHeuristicConfig(),
    ]
    new_problem = problem_setup.model_copy(deep=True)
    new_points: set[PredicateArgument] = set()
    max_new_points = max_new_points or 50
    clauses_per_heuristic: dict[HeuristicName, list[JGEXClauseConsequences]] = {}
    for heuristic_config in complete_the_picture_heuristics_configs:
        heuristic = heuristic_from_config(heuristic_config)
        problem_setup, added_clauses_consequences = apply_heuristics_on_nc_problem(
            problem_setup=problem_setup,
            jgex_problem=jgex_problem,
            heuristic=heuristic,
            rng=rng,
            max_new_points=max_new_points - len(new_points),
        )
        new_problem = problem_setup
        clauses_per_heuristic[heuristic_config.heuristic_name] = (
            added_clauses_consequences
        )
        for clause_consequences in added_clauses_consequences:
            new_points.update(clause_consequences.new_points.keys())

    return new_problem, jgex_problem, clauses_per_heuristic


def apply_heuristics_on_nc_problem(
    problem_setup: ProblemSetup,
    jgex_problem: JGEXFormulation,
    heuristic: Heuristic,
    rng: numpy.random.Generator,
    max_new_points: int | None = None,
) -> tuple[ProblemSetup, list[JGEXClauseConsequences]]:
    goal_less_problem = problem_setup.model_copy(deep=True)
    goal_less_problem.goals = tuple()  # Force saturation
    solver: GeometricSolver = GeometricSolverBuilder(rng=rng).build(goal_less_problem)
    solver.run()

    newclid_problem = problem_setup.model_copy()
    predicate_constructions_deduced = [
        predicate_to_construction(predicate)
        for predicate in solver.proof_state.graph.hyper_graph.keys()
    ]

    goals = problem_setup.goals
    relevant_predicates_constructions = (
        set(goals)
        .union(newclid_problem.assumptions)
        .union(predicate_constructions_deduced)
    )
    points: list[PredicateArgument] = [p.name for p in newclid_problem.points]
    lines, circles, angles, free_points = (
        read_geometric_objects_from_predicate_constructions(
            jgex_problem, relevant_predicates_constructions
        )
    )

    heuristic_setup = HeuristicSetup(
        points=tuple(points),
        free_points=tuple(free_points),
        lines=tuple(lines),
        circles=tuple(circles),
        angles=tuple(angles),
        goals=goals,
    )

    max_new_points = max_new_points or 50
    added_points: set[PredicateArgument] = set()
    new_clauses = heuristic.new_clauses(heuristic_setup, max_new_points, rng)
    successful_clauses_consequences: list[JGEXClauseConsequences] = []
    for clause in new_clauses:
        try:
            newclid_problem, clause_consequences = add_clause_to_problem(
                newclid_problem, clause, defs, rng, 1
            )
        except JGEXConstructionError as e:
            LOGGER.warning(f"Failed to add clause {clause}: {e}")
            continue
        except ValueError as e:
            LOGGER.warning(
                f"Failed to add clause due to a validation error for clause {clause}: {e}"
            )
            continue

        added_points.update(clause_consequences.new_points.keys())
        successful_clauses_consequences.append(clause_consequences)

    return newclid_problem, successful_clauses_consequences


def build_nc_problem_from_jgex_problem(
    jgex_problem: JGEXFormulation,
    rng: numpy.random.Generator,
) -> ProblemSetup:
    goal_less_problem = jgex_problem.model_copy(deep=True)
    jgex_problem_setup = (
        JGEXProblemBuilder(rng=rng).with_problem(goal_less_problem).build()
    )
    solver: GeometricSolver = GeometricSolverBuilder(rng=rng).build(jgex_problem_setup)
    return solver.proof_state.problem


if __name__ == "__main__":
    problem_txt = (
        "a b = segment a b; c = midpoint c a b; d = eqdistance d a a b, eqdistance d b a b;"
        " o = midpoint o c d; e = on_dia e c d; f = on_circum f c d e, eqdistance f e c d"
        " ? coll f o e"
    )
    jgex_problem = JGEXFormulation.from_text(problem_txt)
    nc_problem = build_nc_problem_from_jgex_problem(jgex_problem, rng)
    new_problem, _resulting_jgex_problem, _clauses_consequences_per_heuristic = (
        apply_complete_the_picture_heuristics(
            nc_problem, jgex_problem, rng, max_new_points=20
        )
    )
    new_solver = GeometricSolverBuilder().build(new_problem)

    new_success = new_solver.run()
    print(new_solver.run_infos)



from typing import Sequence

import pytest
from newclid.agent.follow_deductions import (
    CachedDeduction,
    CachedRuleDeduction,
    DeductionProvider,
    DeductionType,
    DoubleCheckError,
    FollowDeductions,
)
from newclid.api import GeometricSolverBuilder
from newclid.justifications.justification import Assumption, justify_dependency
from newclid.numerical.geometries import PointNum
from newclid.predicate_types import PredicateArgument
from newclid.predicates import Predicate, predicate_from_construction
from newclid.problem import PredicateConstruction, ProblemSetup
from newclid.proof_state import ProofState
from newclid.rng import setup_rng
from newclid.rule import Rule
from newclid.symbols.points_registry import Point


def p(name: str) -> PredicateArgument:
    return PredicateArgument(name)


class TestDoubleCheckRuleDeduction:
    def test_successfuly_follows_rule_deduction(self):
        self.fixture.given_cached_deductions(
            [
                CachedRuleDeduction(
                    deduction_type=DeductionType.RULE,
                    rule=self.rule,
                    premises=(
                        self.coll_a_b_c,
                        self.perp_a_b_c_d,
                        self.perp_b_c_a_e,
                    ),
                    conclusions=(self.para_a_e_c_d,),
                    point_deps=[],
                )
            ]
        )

        self.fixture.given_pred_graph_contains_by_construction(
            [self.coll_a_b_c, self.perp_a_b_c_d, self.perp_b_c_a_e]
        )
        self.fixture.when_following_deduction(self.nc_problem)
        self.fixture.then_pred_graph_should_contain(
            conclusion=self.para_a_e_c_d,
            why=(
                self.coll_a_b_c,
                self.perp_a_b_c_d,
                self.perp_b_c_a_e,
            ),
        )

    def test_raise_if_premises_not_in_current_graph(self):
        self.fixture.given_cached_deductions(
            [
                CachedRuleDeduction(
                    deduction_type=DeductionType.RULE,
                    rule=self.rule,
                    premises=(
                        self.coll_a_b_c,
                        self.perp_a_b_c_d,
                        self.perp_b_c_a_e,
                    ),
                    conclusions=(self.para_a_e_c_d,),
                    point_deps=[],
                )
            ]
        )

        self.fixture.given_pred_graph_contains_by_construction(
            [self.coll_a_b_c, self.perp_b_c_a_e]
        )

        with pytest.raises(
            DoubleCheckError, match="Premise 'AB ⟂ CD' .* not found in proof state"
        ):
            self.fixture.when_following_deduction(self.nc_problem)

    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = FollowDeductionFixture()

        self.nc_problem = ProblemSetup(
            name="test-problem-para-from-perps",
            points=(
                Point(name=p("a"), num=PointNum(x=0, y=0)),
                Point(name=p("b"), num=PointNum(x=1, y=0)),
                Point(name=p("c"), num=PointNum(x=2, y=0)),
                Point(name=p("d"), num=PointNum(x=2, y=1)),
                Point(name=p("e"), num=PointNum(x=0, y=1)),
            ),
            assumptions=(),
            goals=(),
        )

        self.coll_a_b_c = PredicateConstruction.from_str("coll a b c")
        self.perp_a_b_c_d = PredicateConstruction.from_str("perp a b c d")
        self.perp_b_c_a_e = PredicateConstruction.from_str("perp b c a e")
        self.para_a_e_c_d = PredicateConstruction.from_str("para a e c d")

        self.rule = Rule(
            id="fake-rule",
            description="A fake rule for testing",
            premises_txt=("coll A B C", "perp A B C D", "perp B C A E"),
            conclusions_txt=("para A E C D",),
        )


class FakeDeductionProvider(DeductionProvider):
    def __init__(self):
        self.cached_deductions: list[CachedDeduction] = []

    def with_cached_deductions(self, deductions: Sequence[CachedDeduction]):
        self.cached_deductions = list(deductions)

    def ordered_deductions_for_problem(
        self, problem: ProblemSetup
    ) -> list[CachedDeduction]:
        return self.cached_deductions

    @property
    def precomputation_input_str(self) -> str:
        return "Fake deduction provider input"


class FollowDeductionFixture:
    def __init__(self):
        self.rng = setup_rng(42)
        self.deductions_provider = FakeDeductionProvider()
        self.predicates_by_construction: list[PredicateConstruction] = []
        self.proof_state: ProofState | None = None

    def given_pred_graph_contains_by_construction(
        self, predicates_by_construction: list[PredicateConstruction]
    ):
        self.predicates_by_construction = predicates_by_construction

    def given_cached_deductions(self, deductions: Sequence[CachedDeduction]):
        self.deductions_provider.with_cached_deductions(deductions)

    def when_following_deduction(self, nc_problem: ProblemSetup):
        agent = FollowDeductions(self.deductions_provider)
        solver = (
            GeometricSolverBuilder(rng=self.rng)
            .with_deductive_agent(agent)
            .build(nc_problem)
        )

        self.proof_state = solver.proof_state
        for predicate_by_construction in self.predicates_by_construction:
            predicate = predicate_from_construction(
                predicate_by_construction,
                points_registry=self.proof_state.symbols.points,
            )
            if predicate is None:
                raise ValueError(
                    f"Statement {predicate_by_construction} could not be built."
                )
            dep = Assumption(predicate=predicate)
            self.proof_state.apply(dep)
        solver.run()

    def then_pred_graph_should_contain(
        self, conclusion: PredicateConstruction, why: tuple[PredicateConstruction, ...]
    ):
        if self.proof_state is None:
            raise ValueError("Proof state is not set, run the solver first.")

        pred_graph = self.proof_state.graph
        conclusion_predicate = predicate_from_construction(
            conclusion, points_registry=self.proof_state.symbols.points
        )
        if conclusion_predicate is None:
            raise ValueError(
                f"Conclusion {conclusion} could not be built from construction."
            )
        dep = pred_graph.hyper_graph.get(conclusion_predicate)
        assert dep is not None

        expected_why_predicates: list[Predicate] = []
        for why_construction in why:
            expected_why_predicate = predicate_from_construction(
                why_construction,
                points_registry=self.proof_state.symbols.points,
            )
            if expected_why_predicate is None:
                raise ValueError(
                    f"Why predicate {why_construction} could not be built from construction."
                )
            expected_why_predicates.append(expected_why_predicate)
        assert set(justify_dependency(dep, self.proof_state)) == set(
            expected_why_predicates
        )



"""Define the proof data outputed by Newclid then used to write and display the proof."""

from __future__ import annotations

from typing import Iterable, NewType, Self

from pydantic import BaseModel, model_validator

from newclid.justifications._index import JustificationType
from newclid.justifications.justification import Justification, justify_dependency
from newclid.predicates import Predicate, predicate_from_construction
from newclid.problem import PredicateConstruction
from newclid.proof_justifications import goals_justifications
from newclid.proof_state import ProofState
from newclid.symbols.points_registry import Point

ProofId = NewType("ProofId", str)


class PredicateInProof(BaseModel):
    """A predicate in the proof with its unique id."""

    id: ProofId
    predicate: Predicate

    def __str__(self) -> str:
        return f"[{self.id}] {str(self.predicate)}"


class ProofStep(BaseModel):
    """A deduction step of the proof."""

    proven_predicate: PredicateInProof
    justification: Justification
    applied_on_predicates: tuple[ProofId, ...]

    def __str__(self) -> str:
        premises_str = f"[{','.join(self.applied_on_predicates)}]"
        return f"{premises_str} =({self.justification.dependency_type.value})> {self.proven_predicate}"


class ProofData(BaseModel):
    proof_length: int = 0
    """Total length of the proof."""
    proof_rules_length: int = 0
    """Number of rule applications in the proof."""
    points: list[Point]
    """Points in the initial problem setup."""
    proven_goals: list[PredicateInProof]
    """Predicate in proof of goals that were proved."""
    unproven_goals: list[Predicate]
    """Goals to prove from the initial problem that were not proved."""
    construction_assumptions: list[PredicateInProof]
    """Assumptions made initialy by construction."""
    numerical_checks: list[PredicateInProof]
    """Assumptions made by numerical checks."""
    trivial_predicates: list[PredicateInProof]
    """Predicates that are considered trivialy true within the proof."""
    proof_steps: list[ProofStep]
    """Proof steps."""

    @model_validator(mode="after")
    def set_proof_length(self) -> Self:
        self.proof_length = len(self.proof_steps)
        self.proof_rules_length = sum(
            1
            for step in self.proof_steps
            if step.justification.dependency_type == JustificationType.RULE_APPLICATION
        )
        return self


def proof_data_from_state(
    goals_constructions: list[PredicateConstruction],
    proof_state: ProofState,
) -> ProofData:
    goals: list[Predicate] = []
    for goal_construction in goals_constructions:
        goal = predicate_from_construction(
            goal_construction, points_registry=proof_state.symbols.points
        )
        if goal is None:
            raise ValueError(f"Could not construct goal {goal_construction}")
        goals.append(goal)

    justifications_for_proven_goals, goal_is_proven = goals_justifications(
        goals, proof_state
    )
    return proof_justifications_to_proof_data(
        proof_justifications=justifications_for_proven_goals,
        proof_state=proof_state,
        goal_is_proven=goal_is_proven,
    )


def proof_justifications_to_proof_data(
    proof_justifications: Iterable[Justification],
    proof_state: ProofState,
    goal_is_proven: dict[Predicate, bool],
) -> ProofData:
    proof_steps: list[ProofStep] = []
    construction_assumptions: dict[Predicate, ProofId] = {}
    numerical_checks: dict[Predicate, ProofId] = {}
    predicates_proven: dict[Predicate, ProofId] = {}
    trivial_predicates: dict[Predicate, ProofId] = {}
    new_predicates: list[Predicate] = []

    for justification in proof_justifications:
        predicate = justification.predicate
        match justification.dependency_type:
            case (
                JustificationType.RULE_APPLICATION
                | JustificationType.AR_DEDUCTION
                | JustificationType.CIRCLE_MERGE
                | JustificationType.LINE_MERGE
                | JustificationType.DIRECT_CONSEQUENCE
            ):
                pass
            case JustificationType.ASSUMPTION:
                tag = _get_or_create_tag(
                    justification.predicate,
                    construction_assumptions,
                    ProofId(f"C{len(construction_assumptions)}"),
                )
                construction_assumptions[predicate] = tag
                predicates_proven[predicate] = tag
                continue
            case JustificationType.NUMERICAL_CHECK:
                tag = _get_or_create_tag(
                    justification.predicate,
                    numerical_checks,
                    ProofId(f"N{len(numerical_checks)}"),
                )
                numerical_checks[predicate] = tag
                predicates_proven[predicate] = tag
                continue
            case JustificationType.REFLEXIVITY:
                tag = _get_or_create_tag(
                    justification.predicate,
                    trivial_predicates,
                    ProofId(f"T{len(trivial_predicates)}"),
                )
                trivial_predicates[predicate] = tag
                predicates_proven[predicate] = tag
                continue

        tag = _get_or_create_tag(
            justification.predicate,
            predicates_proven | numerical_checks | construction_assumptions,
            ProofId(f"{len(new_predicates)}"),
        )
        depends_on: list[ProofId] = []
        for previous_stmt in justify_dependency(justification, proof_state):
            if previous_stmt not in predicates_proven:
                previous_stmt_dep = proof_state.justify(previous_stmt)
                if (
                    previous_stmt_dep.dependency_type
                    == JustificationType.NUMERICAL_CHECK
                ):
                    tag = _get_or_create_tag(
                        previous_stmt,
                        numerical_checks,
                        ProofId(f"N{len(numerical_checks)}"),
                    )
                    numerical_checks[previous_stmt] = tag
                    predicates_proven[previous_stmt] = tag
                    continue

                else:
                    raise KeyError(
                        "Could not find predicate %s in the list of previously proven predicates: %s",
                        previous_stmt,
                        predicates_proven,
                    )
            depends_on.append(predicates_proven[previous_stmt])

        new_predicates.append(predicate)
        predicates_proven[predicate] = tag
        proof_steps.append(
            ProofStep(
                proven_predicate=PredicateInProof(
                    id=tag, predicate=justification.predicate
                ),
                justification=justification,
                applied_on_predicates=tuple(depends_on),
            )
        )

    proven_goals: list[PredicateInProof] = []
    unproven_goals: list[Predicate] = []
    for goal, is_proven in goal_is_proven.items():
        if is_proven:
            proven_goals.append(
                PredicateInProof(id=predicates_proven[goal], predicate=goal)
            )
        else:
            unproven_goals.append(goal)

    return ProofData(
        points=list(proof_state.problem.points),
        proven_goals=proven_goals,
        unproven_goals=unproven_goals,
        construction_assumptions=[
            PredicateInProof(id=tag, predicate=construction)
            for construction, tag in construction_assumptions.items()
        ],
        numerical_checks=[
            PredicateInProof(id=tag, predicate=construction)
            for construction, tag in numerical_checks.items()
        ],
        trivial_predicates=[
            PredicateInProof(id=tag, predicate=construction)
            for construction, tag in trivial_predicates.items()
        ],
        proof_steps=proof_steps,
    )


def _get_or_create_tag(
    predicate: Predicate,
    existing_predicates: dict[Predicate, ProofId],
    new_tag: ProofId,
) -> ProofId:
    if predicate in existing_predicates:
        return existing_predicates[predicate]
    return new_tag



import os
from pathlib import Path

import matplotlib.pyplot as plt

from newclid.draw.figure import draw_figure
from newclid.draw.theme import DrawTheme
from newclid.proof_state import ProofState

human_agent_index: str = r"""
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>SVG and Iframes Layout</title>
    <style>
        body {
            display: flex;
            margin: 0;
            height: 100vh;
        }
        .svg-container {
            flex: 1;
            display: flex;
            justify-content: center;
            align-items: center;
            border-right: 1px solid #ccc;
        }
        .iframe-container {
            flex: 2;
            display: flex;
            flex-direction: column;
            justify-content: space-around;
            padding: 10px;
        }
        iframe {
            width: 100%;
            height: 45%;
            border: 1px solid #ccc;
        }
    </style>
</head>
<body>
    <div class="svg-container">
        <img src="geometry.svg" alt="Geometry SVG">
    </div>
    <div class="iframe-container">
        <iframe src="dependency_graph.html"></iframe>
    </div>
</body>
</html>
"""


def pull_to_server(proof: ProofState, *, server_path: Path):
    os.makedirs(server_path, exist_ok=True)
    with open(server_path / "index.html", "w") as f:
        f.write(human_agent_index)

    fig, ax = plt.subplots()  # pyright: ignore
    draw_figure(
        fig,
        ax,
        proof,
        save_to=server_path / "geometry.svg",
        jgex_problem=None,
        theme=DrawTheme(),
    )
    if proof.check_goals():
        proof.graph.save_pyvis(
            path=server_path / "dependency_graph.html",
            stars=proof.goals,
            proof_state=proof,
        )
    else:
        proof.graph.save_pyvis(
            path=server_path / "dependency_graph.html",
            proof_state=proof,
        )
